# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import time
from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


class Library(object):
    def __init__(self, data):
        self.current_time = int(time.time() * 1000)

        self.version = data["version"]
        self.manifest = data["manifest"]
        self.class_count = data["class_count"]
        self.file_path = data["file_path"]
        self.url = data["url"]
        self.hash_code = data["hash_code"]

    def to_json(self, settings):
        return {
            "classCount": self.class_count,
            "file": self.file_path,
            "hash": self.hash_code,
            "manifest": self.manifest,
            "url": self.url,
            "version": self.version,
            "externalDate": self.current_time,
            "internalDate": self.current_time,
            "tags": settings.config.get("inventory.tags"),
        }



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


class Route(object):
    """
    Route object used for various TS messages
    """

    def __init__(self, verb, url, route):
        self.verb = verb or "GET"
        self.url = url or "/"
        self.signature = route or ""
        self.sources = []

    def to_json_inventory(self):
        """json representation used in v1.0 ApplicationInventory.routes"""
        return {
            "signature": self.signature,
            "verb": self.verb,
            "url": self.url,
        }

    def to_json_traces(self):
        """json representation used in ng Traces.routes"""
        return {
            # "The number of times this route was observed; must be more than 0"
            "count": 1,
            "observations": [{"url": self.url, "verb": self.verb}],
            "signature": self.signature,
        }



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import requests

from .base_ts_message import BaseTsAppMessage
from contrast.utils.decorators import fail_loudly
from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


class ApplicationUpdate(BaseTsAppMessage):
    def __init__(self, libraries):
        super().__init__()

        self.headers.update({"Content-Type": "application/json"})

        # activity message sends "components" aka "architectures"
        # so we will not send the "components" field at this time.

        # field "timestamp" represents the amount of time that has passed
        # since the app settings were changed (not an actual timestamp).
        self.body = {
            "timestamp": self.since_last_update,
            "libraries": [
                lib.to_json(self.settings) for lib in libraries if lib.hash_code
            ],
        }

    @property
    def name(self):
        return "application-update"

    @property
    def path(self):
        return "update/application"

    @property
    def request_method(self):
        return requests.put

    @fail_loudly("Failed to process ApplicationUpdate response")
    def process_response(self, response, reporting_client):
        self.process_response_code(response, reporting_client)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import sys

import contrast
from contrast_vendor.wrapt import register_post_import_hook
from contrast.agent.assess.policy.analysis import analyze
from contrast.agent.policy import patch_manager
from contrast.agent.policy.applicator import apply_assess_patch
from contrast.agent.policy import registry
from contrast.patches import urllib_patch
from contrast.utils.decorators import fail_quietly
from contrast.utils.patch_utils import build_and_apply_patch, wrap_and_watermark
from contrast.agent import scope

from contrast_vendor import structlog as logging


logger = logging.getLogger("contrast")

PYRAMID_SESSION_MODULE = "pyramid.session"
PYRAMID_ROUTER_MODULE = "pyramid.router"


@fail_quietly("Failed to apply policy to new session class")
def _apply_policy(session_cls):
    for patch_policy in registry.get_policy_by_module(PYRAMID_SESSION_MODULE):
        if patch_policy.class_name == "CookieSession":
            apply_assess_patch(session_cls, patch_policy)


@fail_quietly("Failed to apply assess policy for BaseCookieSessionFactory")
def _apply_assess(result, args, kwargs):
    context = contrast.CS__CONTEXT_TRACKER.current()

    patch_policy = registry.get_policy_by_name(
        "pyramid.session.BaseCookieSessionFactory"
    )

    analyze(context, patch_policy, result, args, kwargs)


def build_base_cookie_session_factory_patch(orig_func, _):
    def base_cookie_session_factory(wrapped, instance, args, kwargs):
        """
        base_cookie_session_factory is a function that returns a new CookieSession class

        Since we can't instrument the new class directly using normal policy machinery,
        we instead apply our policy on-demand to the newly created class.
        """
        del instance

        session_cls = None
        try:
            session_cls = wrapped(*args, **kwargs)
            _apply_policy(session_cls)
        finally:
            _apply_assess(session_cls, args, kwargs)

        return session_cls

    return wrap_and_watermark(orig_func, base_cookie_session_factory)


def patch_session_pyramid(pyramid_session_module):
    build_and_apply_patch(
        pyramid_session_module,
        "BaseCookieSessionFactory",
        build_base_cookie_session_factory_patch,
    )


def patch_encode_pyramid(pyramid_encode_module):
    # We go ahead and apply all urllib patches here (even though policy
    # patches will happen later on) because we MUST have some urllib policy patches
    # already applied for these non-policy patches to work.
    # This would not be necessary if in _enable_patches policy_patches were applied
    # first.
    urllib_patch.register_patches()

    import urllib.parse

    # Make sure that we use patched versions of these functions. This is
    # probably handled by repatching in most circumstances but this change was
    # necessary in order to get the pyramid patch tests to pass in the full
    # test suite.
    pyramid_encode_module.quote_plus = urllib.parse.quote_plus
    pyramid_encode_module._url_quote = urllib.parse.quote

    # We can reuse the urllib.parse.urlencode patch since it's exactly the same
    # as the pyramid.encode.urlencode patch
    build_and_apply_patch(
        pyramid_encode_module, "urlencode", urllib_patch.build_urlencode_patch
    )


def build_call_patch(orig_func, patch_policy):
    """
    Patch for pyramid.router.Router.__call__

    This is how we perform route discovery and observation for Pyramid.
    """
    del patch_policy

    def call_patch(wrapped, instance, args, kwargs):
        try:
            result = wrapped(*args, **kwargs)
        finally:
            do_pyramid_routes_discovery(instance)
            do_pyramid_routes_observation(instance, args)
        return result

    return wrap_and_watermark(orig_func, call_patch)


@fail_quietly("unable to perform Pyramid routes discovery")
@scope.with_contrast_scope
def do_pyramid_routes_discovery(pyramid_router_instance):
    from contrast.agent.middlewares.route_coverage.pyramid_routes import (
        create_pyramid_routes,
    )

    from contrast.agent import agent_state

    if not agent_state.is_first_request():
        return

    discovered_routes = create_pyramid_routes(pyramid_router_instance.registry)
    agent_state.update_routes(discovered_routes)

    logger.debug(
        "Discovered the following Pyramid routes: %s",
        [f"{route.verb} {route.url}" for route in discovered_routes.values()],
    )


@fail_quietly("unable to perform Pyramid route observation")
@scope.with_contrast_scope
def do_pyramid_routes_observation(pyramid_router_instance, request_path_arg):
    from contrast.agent.middlewares.route_coverage.pyramid_routes import (
        get_view_func,
        build_route,
    )

    request_path = request_path_arg[0]["PATH_INFO"]
    view_func = get_view_func(
        request_path,
        pyramid_router_instance.routes_mapper.routelist,
        pyramid_router_instance.registry,
    )

    context = contrast.CS__CONTEXT_TRACKER.current()
    if context is None:
        return

    context.view_func = view_func
    context.view_func_str = build_route(
        context.view_func,
        context.request.get_normalized_uri(),
    )
    logger.debug("Observed Pyramid route: %s", context.view_func_str)


def patch_pyramid(pyramid_router_module):
    build_and_apply_patch(pyramid_router_module.Router, "__call__", build_call_patch)


def register_patches():
    register_post_import_hook(patch_pyramid, PYRAMID_ROUTER_MODULE)
    register_post_import_hook(patch_session_pyramid, PYRAMID_SESSION_MODULE)
    register_post_import_hook(patch_encode_pyramid, "pyramid.encode")


def reverse_patches():
    pyramid_router_module = sys.modules.get(PYRAMID_ROUTER_MODULE)
    pyramid_session = sys.modules.get(PYRAMID_SESSION_MODULE)
    pyramid_encode = sys.modules.get("pyramid.encode")

    if pyramid_session:
        patch_manager.reverse_patches_by_owner(pyramid_session)
    if pyramid_encode:
        patch_manager.reverse_patches_by_owner(pyramid_encode)
        urllib_patch.reverse_patches()
    if pyramid_router_module:
        patch_manager.reverse_patches_by_owner(pyramid_router_module.Router)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import os
import time
import threading

from pip._vendor import pkg_resources

from importlib.util import find_spec

from contrast.utils import get_installed_distributions
from contrast.utils.library_reader.utils import (
    normalize_file_name,
    get_top_level_directories,
    get_data,
    get_hash,
    CONTRAST_AGENT_DIST,
    PY_SUFFIX,
    SO_SUFFIX,
)

from contrast.agent import scope
from contrast.api import Library
from contrast.utils.patch_utils import get_loaded_modules
from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


def get_active_library_names_from_pkg():
    library_names = ["contrast"]
    libraries = list(pkg_resources.working_set)
    for library in libraries:
        library_names.append(library.key)
    return library_names


class LibraryReader(object):
    def __init__(self, settings, send_ts_message_func):
        self.settings = settings
        self.send_ts_message_func = send_ts_message_func
        self.analysis_thread = None
        self.installed_distribution_keys = []

        # All packages installed in the site-packages/ directory
        self.installed_dists = get_installed_distributions()

    def start_library_analysis_thread(self, daemon=True):
        self.analysis_thread = threading.Thread(target=self._read_libraries)
        self.analysis_thread.daemon = daemon

        self.analysis_thread.start()

    def join_library_analysis_thread(self):
        if self.analysis_thread:
            self.analysis_thread.join()

    def _read_libraries(self):
        """
        Looks at every library installed in self.installed_dists, then calls search_dist
        on each dist that has a file top_level.txt, which is autogenerated by installation,
        specifying which files/directories are associated with the package

        :exception: IOError if top_level.txt, METADATA, or RECORD cannot be found
        :return: None
        """
        with scope.contrast_scope():
            logger.debug("Analyzing libraries...")

            all_dists = read_dists(self.installed_dists)
            reportable_dists = []

            for dist in all_dists:
                library_name = dist.get("file_path", "")
                if library_name and library_name != CONTRAST_AGENT_DIST:
                    reportable_dists.append(dist)

            self._send_analysis_results_appstart(reportable_dists)
            self._send_analysis_results_files_loaded(reportable_dists)

    def _send_analysis_results_appstart(self, reportable_dists):
        """
        Send library discovery in ApplicationUpdate message.
        :param reportable_dists: A list of dictionaries containing information about
            each distribution.
        """
        if not reportable_dists:
            return

        libraries = [Library(dist) for dist in reportable_dists]

        logger.debug(
            "Discovered libraries: %s",
            [lib.file_path for lib in libraries],
        )

        logger.debug("Sending ApplicationUpdate message with library analysis results")

        # Import here to prevent circular import
        from contrast.reporting import teamserver_messages

        app_update_msg = teamserver_messages.ApplicationUpdate(libraries)
        self.send_ts_message_func(app_update_msg)

    def _send_analysis_results_files_loaded(self, reportable_dists):
        if not reportable_dists:
            return

        from contrast.reporting import teamserver_messages

        msg = teamserver_messages.LibraryUsage(reportable_dists)

        logger.debug("Sending list of modules loaded to Contrast")

        self.send_ts_message_func(msg)


def read_dists(installed_dists):
    results = []

    for dist in installed_dists:
        analysis_result = search_dist(dist)

        if analysis_result:
            results.append(analysis_result)

    return results


def search_dist(dist):
    """
    Searches directories related to dist, gathering relevant statistics and metadata.
    Then, assuming library was loaded, appends that metadata to the self.libraries.

    Created package is added to the output for the process

    :param dist: contains a pkg_resources.DistInfoDistribution object
    referencing the current library being looked for
    :return: A dictionary containing contrast-relevant information about the current distribution
    """
    # __init__.py is required in order to be a package. If there are multiple subpackages
    # we want to make sure we include all __init__.py files as part of the count
    total_files = list()
    total_used_files = set()
    top_level_dirs = get_top_level_directories(dist)

    for directory in top_level_dirs:
        package_path = os.sep.join([dist.location, directory])

        all_dir_files = get_all_files(package_path)
        dir_used_files = search_modules(package_path, directory)

        total_files.extend(all_dir_files)
        total_used_files.update(dir_used_files)

    file_count = len(total_files)
    # Set the package path to be the first directory mentioned in top_level.txt
    name = str(dist).split(" ")[0]

    version, manifest, url = get_data(dist)

    library_hash = get_hash(name, version)

    result = create_package(version, manifest, file_count, name, url, library_hash)
    result["used_files"] = total_used_files

    return result


def create_package(version, manifest, class_count, name, url, sha):
    """
    Generate a json-serializable representation of a package for teamserver

    :param version: version of package
    :param manifest: all metadata
    :param class_count: count of files
    :param name: name of the library
    :param url: homepage of package
    :param sha: sha1 hash of the name(space)version
    :return: dict of package
    """
    current_time = int(time.time() * 1000)

    return {
        "version": version,
        "manifest": manifest,
        "class_count": class_count,
        "file_path": name,
        "url": url,
        "hash_code": sha,
        "external_ms": current_time,
        "internal_ms": current_time,
    }


def search_modules(package_path, directory_name):
    """
    Searches every module to see if it is associated with the current package being looked at.
    If it is, update used file count.

    :param directory_name: name of top level directory (i.e package or subpackage name)
    :param package_path: Path to the package currently being looked for
    :return: The list of used files for this directory, package_found, name, version, manifest,
        and url
    name, version, manifest, and url are empty strings if the package was already found
    """
    used_files = set()
    sys_modules = get_loaded_modules().values()
    top_level_package = directory_name + "."

    for module in sys_modules:
        if (
            module.__name__
            and (module.__name__ + ".").startswith(top_level_package)
            and hasattr(module, "__file__")
            and module.__file__
        ):
            # First we need to check and see if the top-level package (the name up till the
            # first dot) of __name__ is the same as the current directory_name/module name

            # Then, process __file__ (Path to the loaded file)
            # __file__ is optional. If set, this attribute’s value must be a string. The import
            # system may opt to leave __file__ unset if it has no semantic meaning
            # (e.g. a module loaded from a database).

            # using os.path.realpath because in some environments there is a symlink to the
            # lib/python_version_here/site-packages directory. When checking module.__file__
            # the symlink path is used instead of the resolved path. Resolved path is needed
            # because that is what the package_path will point to.
            module_file = os.path.realpath(module.__file__)

            file_path = module_file if module_file.startswith(package_path) else ""
            if len(file_path) > 0:
                used_files.add(normalize_file_name(file_path))

    return used_files


def find_module_file(module_name, module_location):
    """
    If module is found, returns the module file path under the location specified
    by module_location.

    :param module_name: Name of the module
    :param module_location: Path to the module file
    """
    module_file = None

    try:
        # A module spec contains information used to import a module. Such information,
        # origin/location is needed here to get the location of the module found under
        # module_location. PEP 451 contains additional info on this.
        module_spec = find_spec(module_name)
        if hasattr(module_spec, "origin") and module_spec.origin:
            if module_spec.origin.startswith(module_location):
                module_file = module_spec.origin
        elif hasattr(module_spec, "location") and module_spec.location:
            if module_spec.location.startswith(module_location):
                module_file = module_spec.location
    except Exception as e:
        # Raises ModuleNotFoundError (new in 3.6+) instead of AttributeError if package is in
        # fact not a package (i.e. lacks a __path__ attribute). This is something that
        # shouldn't happen so we log it here
        logger.debug(
            "WARNING - Module %s at location %s is not a package: %s",
            module_name,
            module_location,
            e,
        )

    if module_file and os.path.isfile(module_file):
        return module_file

    return None


def get_all_files(package_path):
    """
    Get all .py and .so files in all directories in package_path.

    :param package_path: Path to the current package being looked or a path to the module file.
        If its a module, we need to get the full path to the module file
    since package_path only includes the first portion of the module name.
    :return: list of files
    """
    total_files = []

    if not os.path.isdir(package_path):
        module_name = os.path.basename(package_path)
        module_location = os.path.dirname(package_path)

        module_file = find_module_file(module_name, module_location)
        if module_file:
            total_files.append(module_file)
        # These next 2 cases can occur if __location__ or __origin__ exists as an attr
        # but they don't point to the location of the module
        # This happened for bottle
        elif os.path.exists(package_path + PY_SUFFIX):
            total_files.append(package_path + PY_SUFFIX)

        return total_files

    for _, _, files in os.walk(package_path):
        for f in files:
            if f.endswith((PY_SUFFIX, SO_SUFFIX)):
                total_files.append(f)

    return total_files



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.

# we need this to ensure `agent_lib.modules.*` is available when importing `agent_lib`
import contrast.agent.agent_lib.modules
from contrast.agent.agent_lib.main import call, initialize, update_log_options
from contrast.agent.agent_lib.semantic_analysis import get_index_of_chained_cmd
from contrast.agent.agent_lib.semantic_analysis import does_file_path_bypass_security
from contrast.agent.agent_lib.semantic_analysis import initialize_semantic_analysis
from contrast.agent.agent_lib.semantic_analysis import (
    does_command_contain_dangerous_path,
)
from contrast.agent.agent_lib.input_tracing import (
    evaluate_header_input,
    evaluate_input_by_type,
    initialize_input_tracing,
    check_sql_injection_query,
    check_cmd_injection_query,
    check_method_tampering,
    map_result_and_free_eval_result,
    map_result_and_free_check_query_sink_result,
    DBType,
)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.policy.registry import register_deadzone_nodes


deadzone_nodes = [
    {
        # Prevent recursive propagation when logging in assess
        # We run the risk of recursively logging propagation events inside of log
        # statements. This is because the logging module sometimes uses streams to
        # output logs, and these logging events can inadvertently cause additional
        # log output that we use for debugging within our assess code. We want to
        # prevent this from occurring. Part of the reason for this is that we now
        # instrument stream reads and writes, but we also are aware that StringIO
        # is implemented with a lot of string building under the hood.
        "module": "logging",
        "class_name": "StreamHandler",
        "method_name": "emit",
        "policy_patch": True,
    },
    {
        # Werkzeug's request body parser (used by flask.request.files) causes a lot of
        # propagation. We deadzone this, but we have source nodes / logic for tracking
        # all resulting objects.
        "module": "werkzeug.formparser",
        "class_name": "FormDataParser",
        "method_name": "parse",
    },
    {
        # This method causes potential false positives for unsafe-code-execution.
        # It is presumed safe to deadzone since it is implemented by the framework.
        # The actual issue is caused by the .name attribute but we don't currently
        # support deadzones for properties.
        "module": "werkzeug.exceptions",
        "class_name": "HTTPException",
        "method_name": "get_body",
    },
    {
        # This is django's request body parser. We have source nodes for the objects
        # storing data that is parsed out of the request body here. This deadzone saves
        # us from performing needless propagation without sacrificing any correctness.
        "module": "django.http.request",
        "class_name": "HttpRequest",
        "method_name": "_load_post_and_files",
    },
    {
        # This is DRF's request body parser. See comments for other body parsers.
        "module": "rest_framework.request",
        "class_name": "Request",
        "method_name": "_parse",
    },
    {
        # Built-in request body parser used by at least Bottle and Pyramid. This patch
        # is actually handled explicitly, but the deadzone here is for clarity. We also
        # use cgi.FieldStorage.__init__ as an entrypoint for explicitly patching the
        # relevant attributes of the FieldStorage object itself (as sources).
        "module": "cgi",
        "class_name": "FieldStorage",
        "method_name": "__init__",
        "policy_patch": False,
    },
    {
        # See FieldStorage. It's not essential that we deadzone this, but it keeps
        # MiniFieldStorage and FieldStorage consistent.
        "module": "cgi",
        "class_name": "MiniFieldStorage",
        "method_name": "__init__",
        "policy_patch": False,
    },
    {
        # UUIDs aren't controllable and they're hex encoded. It is essentially
        # impossible for a UUID to trigger a real vulnerability
        "module": "uuid",
        "class_name": "UUID",
        "method_name": "__str__",
    },
    {
        # This currently causes an unvalidated-redirect in the case where the
        # middleware appends a '/' to the request path and redirects
        "module": "django.middleware.common",
        "class_name": "CommonMiddleware",
        "method_name": "process_request",
    },
    {
        # This code is responsible for rendering the django debug page.
        "module": "django.views.debug",
        "class_name": "ExceptionReporter",
        "method_name": "get_traceback_data",
    },
    {
        # Too much propagation / source creation occurs here. It is not necessary for
        # body tracking since we accomplish this with higher-level source nodes
        "module": "quart.wrappers.request",
        "class_name": "Request",
        "method_name": "_load_form_data",
    },
    {
        # Prevents our rewriter from being applied to assertion rewrites in pytest
        "module": "_pytest.assertion.rewrite",
        "class_name": "AssertionRewritingHook",
        "method_name": "exec_module",
    },
]


register_deadzone_nodes(deadzone_nodes)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import inspect

import contrast
from contrast.agent.settings import Settings
from contrast.configuration import CONTRAST_UI_SRC, DEFAULT_VALUE_SRC, ConfigOptionSrc
from contrast.api.attack import Attack, ProtectResponse
from contrast.api.user_input import UserInput, InputType
from contrast.api.sample import Sample
from contrast.agent.protect.rule import ProtectionRule
from contrast.utils.decorators import fail_loudly, fail_quietly
from contrast.utils.decorators import cached_property
from contrast.utils.stack_trace_utils import build_and_clean_stack
from contrast_vendor import structlog as logging
from contrast.utils.loggers.logger import security_log_msg
from contrast.utils.string_utils import ensure_string

logger = logging.getLogger("contrast")


BLOCKING_RULES = frozenset([ProtectionRule.BLOCK, ProtectionRule.BLOCK_AT_PERIMETER])
PREFILTER_RULES = frozenset([ProtectionRule.BLOCK_AT_PERIMETER])
POSTFILTER_RULES = frozenset(
    [ProtectionRule.BLOCK, ProtectionRule.MONITOR, ProtectionRule.PERMIT]
)

PREFILTER = "prefilter"
INFILTER = "infilter"
POSTFILTER = "postfilter"


class BaseRule(object):
    """
    Base rule object that all protection rules will inherit
    """

    RULE_NAME = "base-rule"

    def __init__(self):
        self.settings = Settings()
        self.settings.protect_rules[self.name] = self

    @property
    def name(self):
        return self.RULE_NAME

    @property
    def mode(self):
        """
        Return the mode for this rule based.

        Order of precedence:
        1. Config (contract_security.yaml)
        2. Settings from TS
        3. Default mode

        We have to assert if config_mode is not None because 0 is falsy
        """
        config_mode = self.mode_from_config()

        return config_mode if config_mode is not None else self.mode_from_settings()

    @cached_property
    def config_rule_path_mode(self):
        return f"protect.rules.{self.name}.mode"

    @property
    def config_source_mode(self):
        """
        Returns how the mode for this rule was set
        """
        default_source = ConfigOptionSrc(
            name=self.config_rule_path_mode, source=DEFAULT_VALUE_SRC
        )
        config_mode = self.mode_from_config()

        if config_mode is not None:
            source_node = self.settings.config.get_src(
                self.config_rule_path_mode, default_source
            )

            # If this is a default node we don't want to return so we can check Contrast UI
            if source_node.source != DEFAULT_VALUE_SRC:
                return source_node

        app_settings = self.settings.direct_ts_app_settings

        for definition in app_settings.protection_rules:
            # "name" here is the "id" in ApplicationSettings
            if definition.id == self.name:
                return ConfigOptionSrc(
                    name=f"contrast.{self.config_rule_path_mode}",
                    source=CONTRAST_UI_SRC,
                )

        return default_source

    def mode_from_config(self):
        """
        Retrieve the mode based on the rule name

        Return None if it does no exist in the config which means default to settings
        """
        return self.settings.config.get(self.config_rule_path_mode, None)

    def mode_from_settings(self):
        """
        Retrieve the mode based on the rule name
        """
        app_settings = self.settings.direct_ts_app_settings

        for definition in app_settings.protection_rules:
            # "name" here is the "id" in ApplicationSettings
            if definition.id == self.name:
                return definition.mode
        return ProtectionRule.OFF

    def is_prefilter(self):
        """
        Checks if a rules mode is for prefilter
        """
        return self.enabled and self.mode in PREFILTER_RULES

    def is_postfilter(self):
        """
        Checks if a rules mode is for postfilter
        """
        return self.enabled and self.mode in POSTFILTER_RULES

    def is_blocked(self):
        """
        Checks if a rules mode is for blocking
        """
        return self.enabled and self.mode in BLOCKING_RULES

    @property
    def enabled(self):
        """
        A rule is enabled only if all 3 conditions are met:
        1. protect is enabled in both config and server settings
        2. rule is not in disabled rules list
        3. rule mode is not OFF
        4. an exclusion wasn't applied from Teamserver
        """
        if not self.settings.is_protect_enabled():
            return False

        disabled_rules = self.settings.config.get("protect.rules.disabled_rules", [])
        if disabled_rules and self.name in disabled_rules:
            return False

        req_ctx = contrast.CS__CONTEXT_TRACKER.current()
        if req_ctx is not None and req_ctx.excluded_protect_rules:
            return self.name not in req_ctx.excluded_protect_rules

        return self.mode != ProtectionRule.OFF

    def should_block(self, attack):
        return attack and attack.response == ProtectResponse.BLOCKED

    def excluded(self, exclusions):
        """
        Check if rule is being excluded from evaluation
        :param exclusions:
        :return: True if excluded, else False
        """

        if not exclusions or len(exclusions) == 0:
            return False

        logger.debug("Checking %s exclusion(s) in %s", len(exclusions), self.name)
        return any(ex.match_protect_rule(self.name) for ex in exclusions)

    def protect_excluded_by_code(self):
        """
        Checks if analysis of code needs to be ignored based on code exceptions

        Checks stack for code exceptions
        """
        code_exclusions = self.settings.code_exclusion_matchers
        if not code_exclusions or len(code_exclusions) == 0:
            return False

        current_frame = inspect.currentframe()
        called_frame = inspect.getouterframes(current_frame, 2)
        logger.debug("caller name: %s", called_frame[1][3])
        stack = inspect.stack()

        return any(
            m.match_protect_rule(self.name) and m.match_code(stack)
            for m in code_exclusions
        )

    def prefilter(self):
        """
        Scans the input analysis for the rule and looks for matched attack signatures

        Will throw a SecurityException if a response needs to be blocked
        """
        logger.debug("PROTECT: Prefilter for %s", self.name)

        attack = self.find_attack(analysis_stage=PREFILTER)
        if attack is None or len(attack.samples) == 0:
            return

        self._append_to_context(attack)

        if attack.response == ProtectResponse.BLOCKED_AT_PERIMETER:
            raise contrast.SecurityException(rule_name=self.name)

    def infilter(self, match_string, **kwargs):
        """
        Scans the input analysis for the rule and looks for matched attack signatures. The call to this method may be
        rule specific and include additional context in a args list.
        """
        if self.mode in [ProtectionRule.OFF, ProtectionRule.PERMIT]:
            return

        logger.debug("PROTECT: Infilter for %s", self.name)

        attack = self.find_attack(match_string, analysis_stage=INFILTER, **kwargs)
        if attack is None or len(attack.samples) == 0:
            return

        self._append_to_context(attack)

        if self.should_block(attack):
            raise contrast.SecurityException(rule_name=self.name)

    @fail_loudly("Failed to run protect rule")
    def protect(self, patch_policy, user_input, args, kwargs):
        if not self.enabled:
            return

        if self.skip_protect_analysis(user_input, args, kwargs):
            return

        user_input = self.convert_input(user_input)
        if not user_input:
            return

        self.log_safely(patch_policy.method_name, user_input)

        self.infilter(user_input, **self.infilter_kwargs(user_input, patch_policy))

    def infilter_kwargs(self, user_input, patch_policy):
        return {}

    def skip_protect_analysis(self, user_input, args, kwargs):
        """
        We only want to run protect on user input that is of a type supported
        by the rule.

        Most rules use this implementation, but some override this depending on
        expected user input types.

        :return: Bool if to skip running protect infilter
        """
        if not user_input:
            return True

        if isinstance(user_input, (str, bytes)):
            return False

        logger.debug(
            "WARNING: unknown input type %s for rule %s", type(user_input), self.name
        )

        return True

    def convert_input(self, user_input):
        return ensure_string(user_input)

    def postfilter(self):
        """
        Scans the input analysis for the rule and looks for matched attack signatures

        Appends attacker to the context if a positive evaluation is found
        """
        logger.debug("PROTECT: Postfilter for %s", self.name)

        if self.mode in [ProtectionRule.OFF, ProtectionRule.PERMIT]:
            return

        attack = self.find_attack(analysis_state=POSTFILTER)
        if attack is None or len(attack.samples) == 0:
            return

        self._append_to_context(attack)

        if self.should_block(attack):
            raise contrast.SecurityException(rule_name=self.name)

    def find_attack(self, candidate_string=None, analysis_stage=None, **kwargs):
        """
        Finds the attacker in the original string if present
        """
        if candidate_string is not None:
            logger.debug("Checking for %s in %s", self.name, candidate_string)

        if self.protect_excluded_by_code():
            return None

        evaluations_for_rule = self.evaluations_for_rule()

        attack = None
        for evaluation in evaluations_for_rule:
            if analysis_stage == POSTFILTER and evaluation.attack_count > 0:
                continue

            if candidate_string:
                if candidate_string.find(evaluation.value) == -1:
                    continue

                attack = self.build_attack_with_match(
                    candidate_string, evaluation, attack, **kwargs
                )
            else:
                attack = self.build_attack_without_match(evaluation, attack, **kwargs)

        return attack

    def build_attack_with_match(
        self, candidate_string, evaluation=None, attack=None, **kwargs
    ):
        attack = self.build_or_append_attack(
            evaluation, attack, candidate_string, **kwargs
        )

        if evaluation:
            evaluation.attack_count += 1

        attack.set_response(self.response_from_mode(self.mode))
        self.log_rule_matched(evaluation, attack.response, candidate_string)
        return attack

    def build_attack_without_match(self, evaluation=None, attack=None, **kwargs):
        if self.mode == ProtectionRule.BLOCK_AT_PERIMETER:
            attack = self.build_or_append_attack(evaluation, attack, **kwargs)

            attack.set_response(self.response_from_mode(self.mode))
            self.log_rule_matched(evaluation, attack.response)
        elif evaluation is None or evaluation.attack_count == 0:
            attack = self.build_or_append_attack(evaluation, attack, **kwargs)
            attack.set_response(ProtectResponse.PROBED)
            self.log_rule_probed(evaluation)

        return attack

    def build_or_append_attack(
        self, evaluation, attack=None, candidate_string=None, **kwargs
    ):
        if attack is None:
            attack = self.build_base_attack()

        attack.add_sample(self.build_sample(evaluation, candidate_string, **kwargs))

        return attack

    def build_base_attack(self):
        return Attack(self.name)

    def build_sample(self, evaluation, candidate_string, **kwargs):
        return self.build_base_sample(evaluation)

    def build_user_input(self, evaluation):
        return UserInput(
            input_type=evaluation.input_type,
            key=evaluation.key,
            value=evaluation.value,
            path=evaluation.path,
            matcher_ids=[idx for idx in evaluation.ids],
        )

    def build_base_sample(self, evaluation, prebuilt_stack=None):
        sample = Sample()

        if evaluation:
            sample.set_user_input(self.build_user_input(evaluation))
        else:
            sample.set_user_input(
                UserInput(input_type=InputType.UNKNOWN, key="", value="")
            )

        stack = prebuilt_stack if prebuilt_stack else build_and_clean_stack()
        sample.set_stack(stack)

        return sample

    def log_rule_matched(self, evaluation, response, _=None):
        """
        Logs the exploit for the rule to the security logger
        """
        outcome = response.name.replace("_", " ")

        if evaluation:
            key = str(evaluation.key) if evaluation.key else ""

            rule_message = (
                f"{self.name} - {ensure_string(evaluation.value, errors='replace')}"
            )

            msg = f"The {evaluation.input_type.name} {key} had a value that successfully exploited {ensure_string(rule_message, errors='replace')}"
        else:
            msg = self.effective_attack_message()

        security_log_msg(msg, Settings().app_name, self.name, outcome)

    def log_rule_probed(self, evaluation):
        """
        Logs the probed attack for the rule to the security logger
        """
        outcome = "INEFFECTIVE"

        if evaluation:
            key = str(evaluation.key) if evaluation.key else ""

            rule_message = (
                f"{self.name} - {ensure_string(evaluation.value, errors='replace')}"
            )

            msg = (
                f"The {evaluation.input_type.name} {key} had a value that matched a signature for, but did not"
                f" successfully exploit {ensure_string(rule_message, errors='replace')}"
            )

        else:
            # TODO: PYT-1343
            msg = self.ineffective_attack_message()

        security_log_msg(msg, Settings().app_name, self.name, outcome)

    def effective_attack_message(self):
        return f"An effective attack was detected against {self.name}."

    def ineffective_attack_message(self):
        return f"An ineffective attack was detected against {self.name}."

    def _append_to_context(self, attack):
        context = contrast.CS__CONTEXT_TRACKER.current()
        if context is None:
            # do not remove; this case is not yet well-understood
            logger.debug("WARNING: failed to get request context in _append_to_context")
            return

        context.attacks.append(attack)

    _RESPONSE_MAP = {
        ProtectionRule.MONITOR: ProtectResponse.MONITORED,
        ProtectionRule.BLOCK: ProtectResponse.BLOCKED,
        ProtectionRule.BLOCK_AT_PERIMETER: ProtectResponse.BLOCKED_AT_PERIMETER,
        ProtectionRule.OFF: ProtectResponse.NO_ACTION,
        ProtectionRule.PERMIT: ProtectResponse.NO_ACTION,
    }

    def response_from_mode(self, mode):
        return self._RESPONSE_MAP.get(mode)

    def evaluations_for_rule(self):
        context = contrast.CS__CONTEXT_TRACKER.current()
        if context is None:
            # do not remove; this case is not yet well-understood
            logger.debug(
                "WARNING: failed to get request context in evaluations_for_rule"
            )
            return []

        return [
            evaluation
            for evaluation in context.user_input_analysis
            if evaluation.rule_id == self.RULE_NAME
        ]

    @fail_quietly("Failed to log user input for protect rule")
    def log_safely(self, method_name, user_input):
        """
        Attempt to log user supplied input but do not fail if unable to do so.
        """
        logger.debug(
            "Applying %s rule method %s with user input %s",
            self.name,
            method_name,
            ensure_string(user_input, errors="replace"),
        )



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from datetime import datetime
import json
from pathlib import Path
from os import path
from contrast.reporting.teamserver_messages import BaseTsMessage
from requests.models import Response

from contrast_vendor import structlog as logging
from contrast.utils.decorators import fail_quietly

logger = logging.getLogger("contrast")


def get_message_body(msg):
    if isinstance(msg, BaseTsMessage):
        return msg.body
    if isinstance(msg, Response) and msg.status_code == 200:
        return msg.json()

    return {}


class RequestAudit(object):
    SUB_DIRS = ("requests", "responses")

    def __init__(self, config):
        self.config = config
        self.messages_path = ""

    @fail_quietly("Unable to prepare request_audit dirs")
    def prepare_dirs(self):
        # grab config request audit path, or the default, write /messages dir
        # create requests/response subdirs under message
        parent_path = self.config.get("api.request_audit.path")

        if "messages" in parent_path:
            self.messages_path = parent_path
        else:
            self.messages_path = path.join(parent_path, "messages")

        for sub_dir in self.SUB_DIRS:
            sub_path = path.join(self.messages_path, sub_dir)
            Path(sub_path).mkdir(parents=True, exist_ok=True)

        logger.debug(f"Created request_audit dirs in %s", self.messages_path)

    def audit(self, msg, response):
        if self.config.get("api.request_audit.requests"):
            self.write_data(msg, "requests", msg.name)

        if self.config.get("api.request_audit.responses"):
            self.write_data(response, "responses", msg.name)

    @fail_quietly("Unable to write request audit data")
    def write_data(self, msg, msg_type, msg_name):
        now = datetime.now()
        epoch = now.timestamp()
        day = now.strftime("%Y%m%d")
        time = f"{day}-{epoch}"

        file_name = f"{time}-{msg_name}-teamserver.json"
        file_path = path.join(self.messages_path, msg_type, file_name)

        data = get_message_body(msg)

        # In the future, we'd like to somehow include status code + headers

        with Path(file_path).open("w", encoding="UTF-8") as target:
            json.dump(data, target, indent=1)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.policy import constants
from contrast.agent.assess.assess_exceptions import ContrastAssessException
from contrast.agent.policy.constants import ALL_ARGS, OBJECT, RETURN
from contrast.api.trace_event import TraceEvent

UNDERSCORE = "_"
COMMA = ","
COLON = ":"


class PolicyNode(object):
    def __init__(
        self,
        module="",
        class_name="",
        instance_method=True,
        method_name="",
        source="",
        target="",
        tags=None,
        policy_patch=True,
    ):
        self.module = module
        self.class_name = class_name or ""

        # If no class_name is given, it indicates that this node corresponds to a
        # standalone function, not a method. This means that instance_method is always
        # False. Otherwise we use the given value of instance_method. However, the
        # default for this argument will be True if not given for all policy nodes,
        # which means that only methods that are **not** instance methods must provide
        # it explicitly. This value is used to determine whether or not the first
        # positional argument should be treated as "self" or not by our policy
        # machinery
        self.instance_method = instance_method if class_name else False

        self.method_name = method_name

        self.sources = self.parse_source_or_target(source)
        self.targets = self.parse_source_or_target(target)

        self.ts_valid_source = self.ts_represent(source)
        self.ts_valid_target = self.ts_represent(target)

        self.tags = set()

        if tags:
            self.tags.update(tags)

        # This attribute indicates whether a given policy node corresponds to a patch
        # that should be applied by policy. It defaults to True. However, many of our
        # policy nodes are **not** patched by policy but instead are patched in a
        # different way. For example, nearly all of our string propagators are applied
        # as C hooks. In these cases we still use the policy node for reporting
        # purposes, but we don't want a policy patch.
        self.policy_patch = policy_patch

        self.skip_stacktrace = False
        self.properties = {}

    @property
    def name(self):
        """
        Fully specified unique name for this node

        Includes module name, class name (if applicable) and method name
        """
        return f"{self.location}.{self.method_name}"

    @property
    def location(self):
        """
        Property indicating the fully specified location of the patched method/function

        For a method of a class:
            <module_path>.<class_name>.<method_name>

        For a function:
            <module_path>.<function_name>
        """
        return f"{self.module}{'.' + self.class_name if self.class_name else ''}"

    def add_property(self, name, value):
        if not name or not value:
            return

        self.properties[name] = value

    @property
    def node_class(self):
        return self.__class__.__name__.replace("Node", "")

    @property
    def node_type(self):
        return TraceEvent.EventType.METHOD

    @property
    def version_mismatch(self):
        """
        Indicates whether this node is incompatible with the current Python version

        Previously the finest granularity we have is Py2.7 vs Py3, but eventually we may
        update it to have more specific version specifications.
        """
        pass

    def parse_source_or_target(self, str_rep):
        """
        Given a string source or target, split it into its list representation
        by splitting on ',' and appending the appropriate represntation to the list.

        :param str_rep: string representation of a source or target, such as 'ARG_0, OBJ'
        :return: list representation of source or target, [0, 'OBJ']
        """
        ret = []

        if not str_rep:
            return []

        for item in str_rep.split(COMMA):
            if item == constants.OBJECT:
                ret.append(item)

            elif item == constants.RETURN:
                ret.append(item)

            elif item == constants.ALL_ARGS:
                ret.append(item)

            elif item == constants.ALL_KWARGS:
                ret.append(item)

            # handle ARG_#
            elif item.startswith(constants.ARG + UNDERSCORE):
                arg_num = item.split(UNDERSCORE)[1]
                ret.append(int(arg_num))

            # handle KWARG:name
            elif item.startswith(constants.KWARG + COLON):
                kwarg_name = item.split(COLON)[1]
                ret.append(kwarg_name)

            else:
                return []

        return ret

    def ts_represent(self, str_rep):
        """
        Convert source/target from policy.json into TS-valid version.
        * ARG_# --> P#
        * ARG_0,KWARG:location --> P0,KWARG:location
            TS will ignore everything after P0; kwargs are not directly supported in TS
        * KWARG:location --> P0
            Pure kwarg functions are rare in Python but we must support them. However,
            kwargs are not directly supported in TS so for now we send a fake P0.
        * ALL_ARGS or ALL_KWARGS --> P0
            TS does not currently support sending an arbitrary number of args/kwargs for
            source/target value
        """
        if not str_rep:
            return ""
        if str_rep.startswith(
            (
                constants.KWARG,
                constants.ALL_ARGS,
                constants.ALL_KWARGS,
            )
        ):
            return "P0"

        return str_rep.replace("ARG_", "P")

    def get_matching_first_target(self, self_obj, ret, args, kwargs):
        node_target = self.targets[0]

        if node_target is None:
            return None

        if node_target == ALL_ARGS:
            return args

        if node_target == RETURN:
            return ret

        if node_target == OBJECT:
            return self_obj

        if args and isinstance(node_target, int):
            return args[node_target]

        return kwargs.get(node_target) if kwargs else None

    def validate(self):
        if not self.module:
            raise ContrastAssessException(
                f"{self.node_class} unknown did not have a proper module. Unable to create."
            )

        if not self.method_name:
            raise ContrastAssessException(
                f"{self.node_class} did not have a proper method name. Unable to create."
            )

        self.validate_tags()

    def validate_tags(self):
        if not self.tags:
            return

        for item in self.tags:
            if not (
                item in constants.VALID_TAGS or item in constants.VALID_SOURCE_TAGS
            ):
                raise ContrastAssessException(
                    f"{self.node_class} {self.name} had an invalid tag. {item} is not a known value."
                )

    def _type_to_action(self, sources_or_targets):
        """
        Convert a list of sources or targets into a TS-valid action string.
        """
        if not sources_or_targets:
            return ""

        if len(sources_or_targets) > 1:
            return constants.ALL_TYPE

        item = sources_or_targets[0]

        if item in (constants.ALL_ARGS, constants.ALL_KWARGS):
            return constants.ALL_TYPE

        if item == constants.OBJECT:
            return constants.OBJECT_KEY

        # only target, not source, can be return type
        if item == constants.RETURN:
            return constants.RETURN_KEY

        return constants.ARG_OR_KWARG

    def build_action(self):
        """
        Convert our action, built from our source and target, into
        the TS appropriate action.

        Creation (source nodes) don't have sources (they are the source)
        Trigger (trigger nodes) don't have targets (they are the target)
        Everything else (propagation nodes) are Source2Target
        :return: self.event_action
        """
        if not self.sources:
            event_action = TraceEvent.Action[constants.CREATION_TYPE]
        elif not self.targets:
            event_action = TraceEvent.Action[constants.TRIGGER_TYPE]
        else:
            source = self._type_to_action(self.sources)
            target = self._type_to_action(self.targets)
            str_action = source + constants.TO_MARKER + target

            event_action = TraceEvent.Action[str_action]

        return event_action

    def __repr__(self):
        return f"{self.__class__.__name__} - {self.name}"



# -*- coding: utf-8 -*-
from argparse import ArgumentParser
from collections import OrderedDict
from datetime import datetime
from sys import version_info

from . import hookspy
from .body import create_hook
from .signatures import get_method_signature, fastcall_typedef, fastcall_kwargs_typedef


PY2 = version_info[0] == 2


STRTYPES = ["unicode", "bytes", "bytearray"]
STREAM_TYPES = ["bytesio", "stringio", "iobase"]
TYPES = STRTYPES + STREAM_TYPES


string_methods = [
    "encode",
    "decode",
    "replace",
    "split",
    "rsplit",
    "join",
    "capitalize",
    "casefold",
    "title",
    "center",
    "expandtabs",
    "partition",
    "ljust",
    "lower",
    "lstrip",
    "rjust",
    "rstrip",
    "rpartition",
    "splitlines",
    "strip",
    "swapcase",
    "translate",
    "upper",
    "zfill",
    "formatmap",
    "removeprefix",
    "removesuffix",
]


stream_methods = [
    "read",
    "readline",
    "readlines",
    "read1",
]


string_hook_template = """    ADD_METHOD_HOOK(method_dict, {1});"""
stream_hook_template = """    ADD_STREAM_HOOK(StreamType, {1}, {2});"""
stream_reverse_hook_template = """    REVERSE_STREAM_HOOK(StreamType, {1}, {2});"""


if PY2:
    string_patch_template = """
PyObject *create_{0}_hook_module(PyObject *self, PyObject *arg) {{
    return Py_InitModule("{0}_hooks", methods);
}}
"""
else:
    string_patch_template = """
PyObject *create_{0}_hook_module(PyObject *self, PyObject *arg) {{
    return PyModule_Create(&{0}_module_def);
}}
"""

method_entry_template = (
    '{{"apply_{1}_hook", apply_{0}_{1}_hook, METH_O, "Enable {0}.{1} hook"}}'
)

method_table_template = """
static PyMethodDef methods[] = {{
    {},
    {{NULL, NULL, 0, NULL}},
}};
"""

method_def_template = """
static struct PyModuleDef {0}_module_def = {{
    PyModuleDef_HEAD_INIT,
    "{0}_hooks",
    "methods for hooking {0} methods",
    -1,
    methods,
    NULL,
    NULL,
    NULL,
    NULL
}};
"""

stream_patch_template = """
int patch_{}_methods(PyTypeObject *StreamType) {{

{}

    /* Indicates success */
    return 0;
}}
"""

stream_reverse_template = """
void reverse_{}_methods(PyTypeObject *StreamType) {{

{}

}}
"""

YEAR = datetime.now().year

C_HEADER = """/*
* Copyright © {} Contrast Security, Inc.
* See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
*/
""".format(
    YEAR
)

source_file_header = """{}/* THIS FILE WAS AUTOMATICALLY GENERATED BY HOOKSPY */
/* Python requires its own header to always be included first */
#define PY_SSIZE_T_CLEAN
#include <Python.h>

#include <contrast/assess/patches.h>
#include <contrast/assess/propagate.h>
#include <contrast/assess/scope.h>
#include <contrast/assess/utils.h>

{};
{};
""".format(
    C_HEADER, fastcall_typedef, fastcall_kwargs_typedef
)


pytype_map = {
    "unicode": ("PyUnicode_Type", hookspy.find_unicode_hook),
    "bytearray": ("PyByteArray_Type", hookspy.find_bytearray_hook),
    "bytes": ("PyBytes_Type", hookspy.find_bytes_hook),
    "bytesio": ("", hookspy.find_bytesio_hook),
    "stringio": ("", hookspy.find_stringio_hook),
    "iobase": ("", hookspy.find_iobase_hook),
}


def transform_name(name):
    # This is kind of annoying, but the name of this method in the C-level
    # method table has an underscore.
    return name.replace("formatmap", "format_map")


def find_hooks(spyfunc, methods):
    hook_methods = OrderedDict()
    for name in methods:
        result = spyfunc(transform_name(name))
        if result is None:
            continue
        hook_methods[name] = result
    return hook_methods


def autogen(typename):
    pytype, spyfunc = pytype_map[typename]
    stream_type = typename in STREAM_TYPES
    methods = stream_methods if stream_type else string_methods
    hook_template = stream_hook_template if stream_type else string_hook_template
    patch_template = stream_patch_template if stream_type else string_patch_template

    macros = []
    declarations = []
    hook_funcs = []
    hook_points = []
    create_hook_methods = []
    hook_method_entries = []
    reverse_points = []
    hook_methods = find_hooks(spyfunc, methods)

    for name, (offset, argtype) in hook_methods.items():
        signature = get_method_signature(typename, name, argtype)

        hook_definition, hook_name = create_hook(typename, name, signature, stream_type)
        declarations.append("{} {}_orig;".format(signature, hook_name))
        hook_funcs.append(hook_definition)

        macro_name = "{}_{}_OFFSET".format(typename, name).upper()
        macros.append("#define {} {}".format(macro_name, offset))

        if not stream_type:
            hook_method = "CREATE_HOOK_METHOD({}, {}, {})".format(
                pytype, hook_name, offset
            )
            create_hook_methods.append(hook_method)

            hook_method_entries.append(method_entry_template.format(typename, name))

        hook_points.append(hook_template.format(pytype, hook_name, macro_name))
        reverse_points.append(
            stream_reverse_hook_template.format(pytype, hook_name, macro_name)
        )

    print(source_file_header)
    for m in macros:
        print(m)
    print("\n")
    for d in declarations:
        print(d)
    print("\n")
    for f in hook_funcs:
        print(f)
    print("\n")
    for m in create_hook_methods:
        print(m)
    print("\n")
    if not stream_type:
        print(method_table_template.format(",\n    ".join(hook_method_entries)))
        print("\n")
        if not PY2:
            print(method_def_template.format(typename))
        print("\n")
    print(patch_template.format(typename, "\n".join(hook_points)))

    if stream_type:
        print(stream_reverse_template.format(typename, "\n".join(reverse_points)))


def main():
    p = ArgumentParser("Automatically generate Python type hooks")
    p.add_argument("type", type=str, help="Python type", choices=TYPES)
    args = p.parse_args()

    autogen(args.type)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import re

from contrast.agent.protect.rule.base_rule import BaseRule
from contrast.agent.protect.rule.sql_injection.default_sql_scanner import (
    DefaultSqlScanner,
)
from contrast.agent.protect.rule.sql_injection.mysql_sql_scanner import MysqlSqlScanner
from contrast.agent.protect.rule.sql_injection.postgres_sql_scanner import (
    PostgresSqlScanner,
)
from contrast.agent.protect.rule.sql_injection.sqlite_sql_scanner import (
    SqliteSqlScanner,
)


class SqlInjection(BaseRule):
    """
    SQL Injection Protection rule
    """

    RULE_NAME = "sql-injection"

    def build_attack_with_match(
        self, candidate_string, evaluation=None, attack=None, **kwargs
    ):
        scanner = kwargs.get("scanner", None)
        if scanner is None:
            scanner = self.get_database_scanner(kwargs.get("database"))

        for match in re.finditer(
            re.compile(re.escape(evaluation.value)), candidate_string
        ):
            # TODO: PYT-2847
            # - use the agent-lib implementation here instead of the scanners
            #   - see the ticket for the previous implementation - paste it here
            #   - something about that implementation isn't exactly right, but it's close
            # - remove everything related to `scanner` (including the scanners themselves)
            # - make sure the third arg to `check_sql_injection_query` is right

            # TODO: PYT-2847 remove this call
            last_boundary, boundary = scanner.crosses_boundary(
                candidate_string, match.start(), evaluation.value
            )

            if scanner.NO_BOUNDARY in [last_boundary, boundary]:
                break

            evaluation.attack_count += 1

            kwargs["start_idx"] = match.start()
            kwargs["end_idx"] = match.end()
            kwargs["boundary_overrun_idx"] = boundary
            kwargs["input_boundary_idx"] = last_boundary
            attack = self.build_or_append_attack(
                evaluation, attack, candidate_string, **kwargs
            )

        if attack is not None:
            attack.set_response(self.response_from_mode(self.mode))
            self.log_rule_matched(evaluation, attack.response, candidate_string)

        return attack

    def build_sample(self, evaluation, query, **kwargs):
        sample = self.build_base_sample(evaluation)
        if query is not None:
            sample.details["query"] = query

        if "start_idx" in kwargs:
            sample.details["start"] = int(kwargs["start_idx"])

        if "end_idx" in kwargs:
            sample.details["end"] = int(kwargs["end_idx"])

        if "boundary_overrun_idx" in kwargs:
            sample.details["boundaryOverrunIndex"] = int(kwargs["boundary_overrun_idx"])

        if "input_boundary_idx" in kwargs:
            sample.details["inputBoundaryIndex"] = int(kwargs["input_boundary_idx"])

        return sample

    def get_database_scanner(self, database):
        scanner = DefaultSqlScanner()

        if database == "MySQL":
            scanner = MysqlSqlScanner()
        elif database == "PostgreSQL":
            scanner = PostgresSqlScanner()
        elif database == "SQLite3":
            scanner = SqliteSqlScanner()

        return scanner

    def infilter_kwargs(self, user_input, patch_policy):
        return dict(database=patch_policy.module)

    def skip_protect_analysis(self, user_input, args, kwargs):
        """
        Some sql libraries use special objects (see from sqlalchemy import text)
        so we cannot just check if user_input is falsy.
        """
        if user_input is None:
            return True

        return False

    def convert_input(self, user_input):
        if not isinstance(user_input, (str, bytes)):
            user_input = str(user_input)

        return user_input



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.api import Route
from contrast.agent.middlewares.route_coverage.common import (
    DEFAULT_ROUTE_METHODS,
    get_normalized_uri,
    build_key,
)
from contrast.agent.middlewares.route_coverage.common import build_route

DEFAULT_ROUTE_METHODS = DEFAULT_ROUTE_METHODS + ("HEAD",)


def create_starlette_routes(starlette_router):
    """
    Returns all the routes registered to a Starlette router as a dict.
    :param starlette_app: Starlette router instance (starlette.routing.Router)
    :return: dict {route_id:  api.Route}
    """
    from starlette.routing import Mount

    routes = {}

    for app_route in starlette_router.routes:
        if isinstance(app_route, Mount):
            mnt_routes = create_starlette_routes(app_route)
            routes.update(mnt_routes)
            continue

        view_func = app_route.endpoint

        route = build_route(app_route.name, view_func)
        route_id = str(id(view_func))
        methods = app_route.methods or DEFAULT_ROUTE_METHODS

        for method_type in methods:
            key = build_key(route_id, method_type)
            routes[key] = Route(
                verb=method_type,
                url=get_normalized_uri(str(app_route.name)),
                route=route,
            )

    return routes



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast_vendor import structlog as logging
from contrast.configuration import CONTRAST_UI_SRC, ConfigOptionSrc

logger = logging.getLogger("contrast")


class DisableReaction(object):
    NAME = "DISABLE"
    ENABLE = "enable"
    MESSAGE = "Contrast received instructions to disable itself - Disabling now"

    @staticmethod
    def run(settings):
        logger.warning(DisableReaction.MESSAGE)

        if settings.config:
            src = ConfigOptionSrc(
                name=f"contrast.{DisableReaction.ENABLE}", source=CONTRAST_UI_SRC
            )
            settings.config.put(DisableReaction.ENABLE, False, src=src)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.utils.deprecated_middleware import deprecated_wsgi


BottleMiddleware = deprecated_wsgi("bottle")



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
"""
Implements AST rewriter for Contrast Agent
"""
import importlib.abc
import importlib.machinery
import importlib.util
import sys
import ast
import copy
import operator
import os

# Seems like it is necessary to import this prior to rewriting to avoid some weird partial import state
import tokenize  # pylint: disable=unused-import

from contrast_vendor.wrapt.importer import _ImportHookChainedLoader

# NOTE: It is extremely important to limit the number of imports used by this
# module. It should be restricted to only those imports that are absolutely
# necessary for the operation of the rewriter, and ideally should include only
# built-in or standard library modules that are already imported by the
# interpreter prior to the evaluation of this module. It is *very* important
# that this module does *not* import the core `contrast` package since that
# would introduce a huge number of dependencies that we do not want.
# By limiting the number of dependencies for the rewriter module, we can ensure
# that a minimal number of modules are already imported prior to the
# application of the rewriter, which means we maximize the coverage of our
# rewriter in application and library code.

ENABLE_REWRITER: str = "__CONTRAST_APPLY_REWRITER"
REWRITE_FOR_PYTEST: str = "__CONTRAST_REWRITE_FOR_PYTEST"

_CONTRAST_MODULES = ["contrast", "contrast_vendor", "contrast_rewriter"]
_CONTRAST_OPERATOR_NAME: str = "contrast__operator"


class _ContrastImportHookChainedLoader(_ImportHookChainedLoader):
    def __init__(self, loader) -> None:
        super().__init__(loader)

        if hasattr(loader, "get_code"):
            self.get_code = self.loader.get_code


class DeferredLogger:
    def __init__(self):
        self.messages = []

    def debug(self, message, *args, **kwargs):
        self.messages.append(("debug", message, args, kwargs))


# Can't use our Namespace here since we don't want to import contrast package yet
class rewriter_module:
    logger = DeferredLogger()
    enabled: bool = False
    registry = set()


def _load_module(source, module, filename):
    """
    Convenience method to compile and execute the given module source

    It seems like we do not need any exception handling here since any
    exception that occurs gets handled further up by the import machinery and
    causes it to fall back on the original loader. This is definitely good for
    us since it means that even if we mess up here somehow, it shouldn't
    prevent the original module from being loaded. It will just be loaded
    without our rewrites.
    """
    code = compile(source, filename, "exec", dont_inherit=True)
    exec(code, module.__dict__)


def populate_operator_module(obj):
    obj.setdefault(_CONTRAST_OPERATOR_NAME, operator)


def _is_contrast_module_name(fullname: str) -> bool:
    fullname_split = fullname.split(".")
    if len(fullname_split) == 0:
        return False
    top_level_name = fullname_split[0]
    return top_level_name in _CONTRAST_MODULES


class ContrastMetaPathFinder(importlib.abc.MetaPathFinder):
    def find_spec(self, fullname, path, target=None):
        """
        The finder is in charge of finding a module's "spec". The spec includes import
        machinery metadata about the module - including its name, source file path, and
        the loader, among others.

        Here, we first use importlib's default machinery to get the spec for the module
        about to be imported. The problem with this spec is that it also uses the
        default loader, which isn't what we want. To get around this, we reuse some
        metadata and generate a new spec that points at our loader.
        """
        if _is_contrast_module_name(fullname):
            return None

        spec = importlib.machinery.PathFinder.find_spec(fullname, path)

        if (
            spec is None
            or spec.origin is None
            or not isinstance(spec.loader, importlib.machinery.SourceFileLoader)
        ):
            rewriter_module.logger.debug(
                "Skipping non-source module - fullname=%s, path=%s",
                fullname,
                getattr(spec, "origin", "<unknown>"),
            )
            return None

        new_spec = importlib.util.spec_from_file_location(
            fullname,
            spec.origin,
            loader=ContrastRewriteLoader(fullname, spec.origin),
            submodule_search_locations=spec.submodule_search_locations,
        )
        if new_spec is None:
            return new_spec

        loader = getattr(new_spec, "loader", None)

        if loader and not isinstance(loader, _ContrastImportHookChainedLoader):
            new_spec.loader = _ContrastImportHookChainedLoader(loader)

        rewriter_module.logger.debug(
            "Updated spec for module: fullname=%s, path=%s",
            fullname,
            spec.origin,
        )
        return new_spec


class ContrastRewriteLoader(importlib.machinery.SourceFileLoader):
    def exec_module(self, module) -> None:
        """
        This method is responsible for actually doing the module `exec`-ing. We take
        control of this system and do the following:
        - read the original source file. We require pyc caching to be disabled for this
        - parse the source file into an AST
        - rewrite the AST
        - compile the AST into a code object
        - exec the code object

        Note that we add our custom add function to the module's globals. This prevents
        the need for import rewriting entirely
        """
        original_source_code = None
        filename = self.path

        # May be None in some cases such as for namespace packages
        if filename is None:
            return

        try:
            original_source_code = self.get_source(self.name)
            tree = ast.parse(original_source_code)
        except Exception as ex:
            rewriter_module.logger.debug(
                "WARNING: failed to rewrite %s", filename, exc_info=ex
            )

            _load_module(original_source_code, module, filename)

            return

        if _CONTRAST_OPERATOR_NAME not in module.__dict__:
            try:
                populate_operator_module(module.__dict__)
                PropagationRewriter().visit(tree)
                ast.fix_missing_locations(tree)
            except Exception as ex:
                rewriter_module.logger.debug(
                    "WARNING: failed to rewrite %s", filename, exc_info=ex
                )
        else:
            rewriter_module.logger.debug(
                "WARNING: module %s appears to have been already rewritten; will not rewrite again",
                filename,
            )

        rewriter_module.registry.add(self.name)

        _load_module(tree, module, filename)


class PropagationRewriter(ast.NodeTransformer):
    def _copy_with_context(self, node, context):
        node = copy.copy(node)
        node.ctx = context
        return node

    def _make_attr(self, op):
        return ast.Attribute(
            value=ast.Name(id=_CONTRAST_OPERATOR_NAME, ctx=ast.Load()),
            attr=op.__name__,
            ctx=ast.Load(),
        )

    def visit_BinOp(self, binop: ast.BinOp):
        """
        If we see an "Add" or a "Mod" binary operation, replace it with a call to our custom add/modulo
        function, which includes all necessary instrumentation.
        """
        binop.left = self.visit(binop.left)
        binop.right = self.visit(binop.right)

        if isinstance(binop.op, ast.Mod):
            binop_replacement = ast.Call(
                func=self._make_attr(operator.mod),
                args=[binop.left, binop.right],
                keywords=[],
            )

            ast.copy_location(binop_replacement, binop)
            return binop_replacement

        if not isinstance(binop.op, ast.Add):
            return binop

        binop_replacement = ast.Call(
            func=self._make_attr(operator.add),
            args=[binop.left, binop.right],
            keywords=[],
        )
        ast.copy_location(binop_replacement, binop)
        return binop_replacement

    def visit_AugAssign(self, node: ast.AugAssign):
        """
        If we see an "Append", `+=` operation, rewrite it as a `+`.
        """
        node.value = self.visit(node.value)

        if not isinstance(node.op, ast.Add):
            return node

        target = left = None
        if isinstance(node.target, ast.Name):
            name = ast.Name(id=node.target.id)
            target = self._copy_with_context(name, ast.Store())
            left = self._copy_with_context(name, ast.Load())
        else:
            target = node.target
            left = self._copy_with_context(target, ast.Load())

        call_contrast_append_node = ast.Assign(
            targets=[target],
            value=ast.Call(
                func=self._make_attr(operator.iadd),
                args=[self.visit(left), node.value],
                keywords=[],
            ),
        )

        ast.copy_location(call_contrast_append_node, node)
        # Recurse here in order to properly rewrite the `+` operation
        return call_contrast_append_node

    def visit_JoinedStr(self, node: ast.JoinedStr):
        return ast.Call(
            func=ast.Attribute(value=ast.Constant(""), attr="join", ctx=ast.Load()),
            args=[ast.List(elts=node.values, ctx=ast.Load())],
            keywords=[],
        )


def _non_contrast_module_filter(module_item) -> bool:
    return not _is_contrast_module_name(
        getattr(module_item[1], "__package__", "") or ""
    )


def _log_imported_modules() -> None:
    module_map = {
        importlib.machinery.BuiltinImporter: "builtin",
        importlib.machinery.FrozenImporter: "frozen",
    }

    all_modules = list(filter(_non_contrast_module_filter, sys.modules.items()))

    rewriter_module.logger.debug(
        "the following %d modules are already imported", len(all_modules)
    )
    for name, module in sorted(all_modules):
        module_type = module_map.get(getattr(module, "__loader__", None), "source")
        rewriter_module.logger.debug("%-20s type=%s", name, module_type)


def _find_pytest_rewriter_index() -> int:
    """
    Find the index of the pytest assertion rewriter

    When we run with Pytest, we want to insert our rewriter *after* the assertion rewriter
    """
    from _pytest.assertion.rewrite import AssertionRewritingHook

    for i, finder in enumerate(sys.meta_path):
        if isinstance(finder, AssertionRewritingHook):
            return i

    # Caller will increment with +1, so just return -1
    return -1


def _hook_assertion_rewrites(rewrite_module):
    from contrast_vendor.wrapt import function_wrapper

    # This hook enables us to apply our rewriter before assertion rewrites are applied
    def rewrite_asserts(wrapped, _, args, kwargs):
        mod = args[0]
        PropagationRewriter().visit(mod)
        ast.fix_missing_locations(mod)
        wrapped(*args, **kwargs)

    # This hook ensures that we add our contrast-specfic functions to the rewritten module
    def exec_module(wrapped, _, args, kwargs):
        module = args[0]
        populate_operator_module(module.__dict__)
        wrapped(*args, **kwargs)

    rewrite_module.rewrite_asserts = function_wrapper(rewrite_asserts)(
        rewrite_module.rewrite_asserts
    )
    rewrite_module.AssertionRewritingHook.exec_module = function_wrapper(exec_module)(
        rewrite_module.AssertionRewritingHook.exec_module
    )


def register_assertion_rewrite_hooks():
    """
    Register hooks for pytest's assertion rewriter

    This is only to be used for internal testing purposes. It enables our
    rewrites to be compatible with pytest's assertion rewrites.
    """
    from contrast_vendor.wrapt import register_post_import_hook

    register_post_import_hook(_hook_assertion_rewrites, "_pytest.assertion.rewrite")


def register(override_config=False, with_pytest=False):
    """
    Register our rewriter with the import system. After this call, any newly imported
    modules (from source code) will use our custom rewriter.

    Note that because this function is defined in the same module that defines our add
    replacement function, we never have to worry about rewriting the addition in the
    replacement function itself. If that were to occur, we would get an infinite
    recursion.

    Rewriter should only run in >=py3.10 and only in environments in which our default
    patching mechanism does not work.

    :param override_config: Force the rewriter to be registered regardless of configuration (for testing purposes)
    :param with_pytest: Indicates whether rewriter is being used in a pytest context (for internal use)
    """
    if not (override_config or os.environ.get(ENABLE_REWRITER) or rewrite_for_pytest()):
        rewriter_module.logger.debug("Rewriter not enabled, exiting")
        return

    if is_rewriter_enabled():
        rewriter_module.logger.debug("Rewriter already enabled, not applying again")
        return

    _log_imported_modules()

    # When we're running our own tests with pytest, we want our rewriter to be
    # inserted *after* pytest's assertion rewriting hook. This means that
    # pytest gets to rewrite assertions for any test modules, and we get to
    # rewrite any other non-test modules that are loaded (since the pytest
    # rewriting hook will defer to the next loader in those cases). We also
    # hook pytest's assertion rewriter so that we can apply our own rewriter to
    # test modules too.
    insert_idx = _find_pytest_rewriter_index() + 1 if with_pytest else 0

    sys.meta_path.insert(insert_idx, ContrastMetaPathFinder())

    rewriter_module.logger.debug("enabled AST rewriter")

    rewriter_module.enabled = True


def deregister():
    """
    Remove our rewriter from the import system. Modules that were loaded by our rewriter
    will remain rewritten.

    Return True if we find and deregister our machinery, False otherwise.
    """
    for i, finder in enumerate(sys.meta_path.copy()):
        if isinstance(finder, ContrastMetaPathFinder):
            sys.meta_path.pop(i)
            rewriter_module.enabled = False
            return True
    return False


def process_rewriter_logs(real_logger):
    if not isinstance(rewriter_module.logger, DeferredLogger):
        real_logger.debug("WARNING: cannot process deferred rewriter logs")
        return

    for level, message, args, kwargs in rewriter_module.logger.messages:
        getattr(real_logger, level)(message, *args, **kwargs)

    rewriter_module.logger.messages.clear()


def set_rewriter_logger(logger):
    rewriter_module.logger = logger


def is_rewriter_enabled() -> bool:
    return rewriter_module.enabled


def rewrite_for_pytest() -> bool:
    return REWRITE_FOR_PYTEST in os.environ


def module_was_rewritten(name: str) -> bool:
    return name in rewriter_module.registry



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.policy.registry import register_source_nodes
from contrast.agent.policy.utils import CompositeNode


falcon_sources = [
    CompositeNode(
        {
            "module": "falcon.request",
            "class_name": "Request",
        },
        [
            {
                "method_name": ["get_param", "params"],
                "node_type": "QUERYSTRING",
                "tags": ["NO_NEWLINES", "CROSS_SITE"],
            },
            {
                "method_name": "get_media",
                "node_type": "BODY",
                "tags": ["NO_NEWLINES", "CROSS_SITE"],
            },
        ],
    ),
    CompositeNode(
        {
            "module": "falcon.asgi.request",
            "class_name": "Request",
        },
        [
            {
                "method_name": "get_param",
                "node_type": "QUERYSTRING",
                "tags": ["NO_NEWLINES", "CROSS_SITE"],
            },
            {
                "method_name": "get_media",
                "node_type": "BODY",
                "tags": ["NO_NEWLINES", "CROSS_SITE"],
            },
        ],
    ),
    {
        "module": "falcon.asgi.stream",
        "class_name": "BoundedStream",
        "method_name": "read",
        "node_type": "BODY",
        "tags": ["NO_NEWLINES", "CROSS_SITE"],
    },
    {
        "module": "falcon.asgi.reader",
        "class_name": "BufferedReader",
        "method_name": "read",
        "node_type": "BODY",
        "tags": ["NO_NEWLINES", "CROSS_SITE"],
    },
]


register_source_nodes(falcon_sources)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.assess.policy.propagators.base_propagator import (
    BasePropagator,
    SUPPORTED_TYPES,
)
from contrast.agent.assess.utils import (
    is_tracked,
    get_properties,
    track_string,
)
from contrast.utils.assess.duck_utils import is_iterable, safe_iterator


class JsonPropagator(BasePropagator):
    """
    JSON Propagator to check if any sources in a dict are tracked
    If tracked, splat tags to result string

    This propagator is specifically for json.dumps; this method will convert a dict to a string
    We need to check if any tracked items in the dict are tracked, if any of them are, we need
    to transition any tags over to the result and track the new dumped string

    Example:
        tracked = 'vuln'
        dict = {'user': 'bob', 'account': tracked}

        result = json.dumps(dict)  # '{"user": "bob", "account": "vuln"}' <= track this
    """

    @property
    def inputs_require_propagation(self):
        return True

    def propagate(self):
        """
        json.dumps only has one source ARG_0;
        ARG_1 for json.dump is the output IO so not a source
        """
        self.first_source = self.sources[0]
        self._propagate(self.first_source)

    def _propagate(self, item):
        if not item:
            return

        if isinstance(item, SUPPORTED_TYPES):
            if is_tracked(item):
                self._apply_tags(item)
        elif isinstance(item, dict):
            for key, value in item.items():
                self._propagate(key)
                self._propagate(value)
        elif is_iterable(item):
            for it in safe_iterator(item):
                self._propagate(it)

    def _apply_tags(self, item):
        """
        Transfer all items from the item to the target, our result
        :param item: value from the dict

        This is a private method and does not overload the base `apply_tags`
        method as it doesn't require the same args.
        """
        target = self.target

        item_properties = get_properties(item)
        target_properties = get_properties(target) or track_string(target)

        if item_properties and target_properties:
            for event in item_properties.events:
                target_properties.events.append(event)

            self.splat_tags(item, target)

            target_properties.cleanup_tags()

        return target_properties

    def track_target(self):
        """NOP; Let propagate handle the decision to track"""
        pass



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import re


from enum import Enum, auto

from contrast.utils.decorators import fail_loudly
from contrast.api import TypeCheckedProperty


from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


class ExclusionType(Enum):
    URL_EXCLUSION_TYPE = auto()
    INPUT_EXCLUSION_TYPE = auto()


class InputExclusionSourceType(Enum):
    COOKIE = auto()
    PARAMETER = auto()
    HEADER = auto()
    BODY = auto()
    QUERYSTRING = auto()
    UNKNOWN = auto()


class TsExclusion:
    type = TypeCheckedProperty(
        ExclusionType, constructor_arg=ExclusionType.URL_EXCLUSION_TYPE
    )
    input_type = TypeCheckedProperty(
        InputExclusionSourceType, constructor_arg=InputExclusionSourceType.UNKNOWN
    )

    def __init__(
        self,
        exclusion_type,
        name,
        modes,
        match_strategy,
        protection_rules,
        assessment_rules,
        urls=None,
        input_type=None,
        input_name=None,
    ):
        self.type = exclusion_type
        self.name = name
        self.modes = modes
        self.match_strategy = match_strategy
        self.protection_rules = protection_rules
        self.assessment_rules = assessment_rules
        self.urls = urls
        self.input_type = getattr(
            InputExclusionSourceType, input_type or "", InputExclusionSourceType.UNKNOWN
        )
        self.input_name = input_name


def strs_to_regexes(patterns, ignorecase=0):
    """

    :param patterns: list of strings
    :return: list of regular expressions
    """
    if not patterns:
        return []

    regexes = []

    for pattern in patterns:
        try:
            pat = re.compile(pattern, flags=ignorecase)
            regexes.append(pat)
        except Exception as ex:
            logger.warning("Unable to parse pattern %s: %s", pattern, ex)

    return regexes


class Exclusions:
    """
    Container for all exclusions
    """

    @staticmethod
    def init_input_exclusions_container():
        input_exclusions = dict(
            HEADER=[], HEADER_KEY=[], PARAMETER=[], COOKIE=[], BODY=[], QUERYSTRING=[]
        )

        # Aliases for the same source type
        input_exclusions["HEADER"] = input_exclusions["HEADER_KEY"]
        input_exclusions["MULTIPART_FORM_DATA"] = input_exclusions["BODY"]
        input_exclusions["MULTIPART_CONTENT_DATA"] = input_exclusions["BODY"]

        return input_exclusions

    def __init__(self, exclusions):
        self.input_exclusions = self.init_input_exclusions_container()
        self.url_exclusions = []

        has_named_exclusion = False

        for exc in exclusions:
            if exc.type == ExclusionType.INPUT_EXCLUSION_TYPE:
                self.input_exclusions[exc.input_type.name].append(InputExclusion(exc))
                has_named_exclusion = True
            else:
                self.url_exclusions.append(UrlExclusion(exc))

        if not has_named_exclusion:
            self.input_exclusions = None

    def evaluate_assess_trigger_time_exclusions(self, context, finding):
        # returns True if we do not report finding
        if not context.input_exclusions_trigger_time:
            return False

        for exc in context.input_exclusions_trigger_time:
            if exc.match_in_finding(finding):
                return True

        return False

    def evaluate_input_exclusions(self, context, source_type, source_name, mode=None):
        # Evaluate all exclusions against the current source
        if context.input_exclusions is None:
            return False

        exclusions = context.input_exclusions.get(source_type, None)
        if exclusions is None:
            return False

        for exc in exclusions:
            if mode and mode not in exc.modes:
                continue

            if exc.match(
                context,
                source_type=source_type,
                source_name=source_name,
            ):
                logger.debug(
                    "The input exclusion rule named '%s' matched on the input name '%s' for the input type of '%s'",
                    exc.exclusion_name,
                    source_name,
                    exc.input_type,
                )
                return True

        return False

    def evaluate_input_exclusions_url(self, context, source_type, path, mode=None):
        # Evaluate all exclusions against the current source
        if context.input_exclusions is None:
            return False

        exclusions = context.input_exclusions.get(source_type, None)
        if exclusions is None:
            return False

        for exc in exclusions:
            if mode and mode not in exc.modes:
                continue

            if exc.match_type == "ALL":
                logger.debug(
                    "The input url exclusion rule named '%s' matched on the path '%s' for the input type of '%s'",
                    exc.exclusion_name,
                    path,
                    exc.input_type,
                )
                return True

            for url_regex in exc.url_regexes:
                if url_regex.search(path):
                    logger.debug(
                        "The input url exclusion rule named '%s' matched on the path '%s' for the input type of '%s'",
                        exc.exclusion_name,
                        path,
                        exc.input_type,
                    )
                    return True

        return False

    def evaluate_url_exclusions(self, context, target):
        """
        This function evaluates all exclusions depending on the request URL and updates the request context to contain
        the list of disabled assess and protect rules to be evaluated at trigger time for url exclusions

        @param context: request context we are evaluating exclusions in
        @param target: The target string to be matched against
        @type target:
        @return: boolean indicating whether or not request should not be analyzed
        @rtype: bool
        """
        has_match = False

        for exc in self.url_exclusions:
            if exc.match(context, target):
                has_match = True

        if has_match:
            from contrast.agent.settings import Settings

            if Settings().is_assess_enabled() and context.excluded_assess_rules is None:
                return True

            if (
                Settings().is_protect_enabled()
                and context.excluded_protect_rules is None
            ):
                return True

        return False

    def set_input_exclusions_by_url(self, context, path):
        """
        Evaluates the set of input exclusions that apply to this path. Update request
        context with input exclusions to apply
        """
        context.input_exclusions = self.init_input_exclusions_container()
        context.input_exclusions_trigger_time = []

        has_match = False

        # pylint: disable=too-many-nested-blocks
        for input_type, exclusions in self.input_exclusions.items():
            for exc in exclusions:
                if exc.url_regexes:
                    for pattern in exc.url_regexes:
                        if pattern.fullmatch(path):
                            has_match = True

                            if exc.protect_rules or exc.assess_rules:
                                context.input_exclusions_trigger_time.append(exc)
                            else:
                                context.input_exclusions[input_type].append(exc)

                            logger.debug(
                                "Path %s matched on input exclusion pattern %s",
                                path,
                                pattern,
                            )

                else:
                    if exc.protect_rules or exc.assess_rules:
                        context.input_exclusions_trigger_time.append(exc)
                    else:
                        context.input_exclusions[input_type].append(exc)

                    has_match = True

        if not has_match:
            # No exclusions for this request
            context.input_exclusions = None


class BaseExclusion:
    def __init__(self, exclusion):
        self.protect_rules = None
        self.assess_rules = None
        self.url_regexes = []
        self.exclusion_name = exclusion.name

        if exclusion.urls:
            self.url_regexes = strs_to_regexes(exclusion.urls)

        self.protect_rules = exclusion.protection_rules
        self.assess_rules = exclusion.assessment_rules

        # TODO: PYT-2613 remove?
        # old comment:
        # This will add the needed match strategy for direct TS requests and leave it compatible with SR
        if hasattr(exclusion, "match_strategy"):
            self.match_type = exclusion.match_strategy
        if hasattr(exclusion, "modes"):
            self.modes = exclusion.modes

    def match(self, context, **kwargs):
        raise NotImplementedError


class InputExclusion(BaseExclusion):
    def __init__(self, exclusion):
        super().__init__(exclusion)

        self.input_name_regex = None

        self.input_type = exclusion.input_type

        if self.input_type == InputExclusionSourceType.UNKNOWN:
            logger.error("Invalid input exclusion type for the exclusion %s", exclusion)

        if exclusion.input_name:
            # Adding a check for type cookie to ignore case as per requirements
            # https://github.com/Contrast-Security-Inc/platform-specifications/blob/main/exclusions/EXCLUSIONS_INPUT.md
            if exclusion.input_type == InputExclusionSourceType.COOKIE:
                self.input_name_regex = strs_to_regexes(
                    [exclusion.input_name], re.IGNORECASE
                )[0]
            else:
                self.input_name_regex = strs_to_regexes([exclusion.input_name])[0]

    def is_body_input_type(self, source_type):
        return self.input_type == InputExclusionSourceType.BODY or source_type in [
            "BODY",
            "MULTIPART_FORM_DATA",
            "MULTIPART_CONTENT_DATA",
        ]

    def is_querystring_input_type(self, source_type):
        return (
            self.input_type == InputExclusionSourceType.QUERYSTRING
            or source_type == "QUERYSTRING"
        )

    def match_in_finding(self, finding):
        creation_action = 0

        # pylint: disable=too-many-nested-blocks
        if finding.rule_id in self.assess_rules:
            for event in finding.events:
                if event.action == creation_action:
                    for trace_event in event.event_sources:
                        event_type = trace_event.type
                        event_src_name = trace_event.name

                        if event_type == self.input_type.name:
                            exclude = False

                            if (
                                self.input_name_regex is not None
                                and self.input_name_regex.fullmatch(event_src_name)
                            ):
                                exclude = True

                            if self.input_type in [
                                InputExclusionSourceType.QUERYSTRING,
                                InputExclusionSourceType.BODY,
                            ]:
                                exclude = True

                            if exclude:
                                logger.debug(
                                    "The input exclusion rule named '%s' matched on the input name '%s' for the input type of '%s' for the rule '%s'",
                                    self.exclusion_name,
                                    event_src_name,
                                    event_type,
                                    finding.rule_id,
                                )
                                return exclude

        return False

    def match(self, context, source_type=None, source_name=None):
        if source_name is None and self.input_type not in [
            InputExclusionSourceType.QUERYSTRING,
            InputExclusionSourceType.BODY,
        ]:
            return False

        if self.is_body_input_type(source_type) or self.is_querystring_input_type(
            source_type
        ):
            return True

        return self.input_name_regex.fullmatch(source_name) is not None


class UrlExclusion(BaseExclusion):
    @fail_loudly("Unable to ignore request", return_value=False)
    def match(self, context, path=None):
        """
        Determine if the given path exactly matches any of the
        configured urls for this exclusion rule. This function modifies the
        request context to notify rules if they should be disabled for this url.

        @param context: request context
        @param path: path for current request
        @return This function returns True if a match was found
        @rtype: bool
        """
        has_match = False

        if path is None:
            return False

        for pattern in self.url_regexes:
            if pattern.fullmatch(path):
                logger.debug("Path %s matched on pattern %s", path, pattern)

                self.update_disabled_rules(context)

                has_match = True

        return has_match

    def update_disabled_rules(self, request_ctx):
        if request_ctx is not None:
            # The spec states that if either assess_rules or protect_rules is None
            # we ignore the full request so set the disabled rules to None to ignore everything
            if request_ctx.excluded_assess_rules is not None and self.assess_rules:
                request_ctx.excluded_assess_rules.extend(self.assess_rules)
            elif not self.assess_rules:
                request_ctx.excluded_assess_rules = None

            if request_ctx.excluded_protect_rules is not None and self.protect_rules:
                request_ctx.excluded_protect_rules.extend(self.protect_rules)
            elif not self.protect_rules:
                request_ctx.excluded_protect_rules = None



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import threading
import queue

from .teamserver_messages import BaseTsMessage
from contrast.agent import scope
from contrast.agent.disable_reaction import DisableReaction
from contrast.agent.settings import Settings
from contrast.reporting import RequestAudit
from contrast_vendor import structlog as logging
from contrast.utils.singleton import Singleton

logger = logging.getLogger("contrast")

REPORTING_CLIENT_THREAD_NAME = "ContrastReportingClient"

MAX_ATTEMPTS = 2
ERROR_STATUS_CODE = -1


class ReportingClient(Singleton, threading.Thread):
    def init(self):
        self.stopped = False
        self.message_q = queue.Queue(maxsize=128)
        self.settings = Settings()
        self.request_audit = (
            RequestAudit(self.settings.config)
            if self.settings.config.is_request_audit_enabled
            else None
        )

        if self.request_audit:
            self.request_audit.prepare_dirs()

        super().__init__()
        # A thread must have had __init__ called, but not start, to set daemon
        self.daemon = True
        self.name = REPORTING_CLIENT_THREAD_NAME

    def start(self):
        self.stopped = False
        super().start()

    def run(self):
        with scope.contrast_scope():
            logger.debug("Starting reporting thread")

            while not self.stopped and self.settings.is_agent_config_enabled():
                try:
                    msg = self.message_q.get(block=True, timeout=5)
                    response = self.send_message(msg)
                    msg.process_response(response, self)
                except queue.Empty:
                    pass
                except Exception as e:
                    logger.debug(
                        "WARNING: reporting client failed to send message", exc_info=e
                    )

    def send_message(self, msg):
        status_code = ERROR_STATUS_CODE
        msg_name = msg.class_name

        try:
            logger.debug("Sending %s message to Teamserver", msg_name)

            cert = None
            url = msg.base_url + msg.path

            if (
                self.settings.is_cert_verification_enabled
                and not self.settings.ignore_cert_errors
            ):
                if (
                    not self.settings.ca_file
                    or not self.settings.client_cert_file
                    or not self.settings.client_private_key
                ):
                    logger.error(
                        "Unable to communicate with Contrast. "
                        "Certificate configuration is not set properly."
                    )
                    DisableReaction.run(self.settings)
                    return None

                cert = (
                    self.settings.client_cert_file,
                    self.settings.client_private_key,
                )

            response = msg.request_method(
                url,
                json=msg.body,
                proxies=msg.proxy,
                headers=msg.headers,
                allow_redirects=False,
                verify=False
                if self.settings.ignore_cert_errors
                else (self.settings.ca_file or True),
                cert=cert,
            )

            try:
                status_code = response.status_code
                msg_success_status = response.json().get("success")
                messages = response.json().get("messages")
                if not msg_success_status:
                    logger.error(
                        "Failure on Contrast UI processing request reason - (%s): %s",
                        messages,
                        status_code,
                    )
            except Exception as e:
                if status_code == ERROR_STATUS_CODE:
                    logger.debug(
                        "Failed to receive response from Contrast UI: %s ",
                        e,
                    )

            logger.debug("Contrast UI response (%s): %s", msg_name, status_code)

            if self.request_audit:
                self.request_audit.audit(msg, response)

            msg.sent()
            return response

        except Exception as e:
            logger.exception("Failed to send %s message to Contrast: %s", msg_name, e)

        return None

    def add_message(self, msg):
        if msg is None or not isinstance(msg, BaseTsMessage):
            return

        logger.debug("Adding msg to reporting queue: %s", msg.class_name)

        self.message_q.put(msg)

    def retry_message(self, msg):
        # Never send a message more than twice (original time plus one retry)
        # To prevent queue from filling up or causing memory issues.
        if msg.sent_count < MAX_ATTEMPTS:
            logger.debug("Re-enqueuing %s message", msg.class_name)
            self.add_message(msg)



from .signatures import (
    fastcall_name,
    fastcall_kwargs_name,
    calls_by_name,
    propagate_args_by_name,
    signatures_by_name,
)


hook_template = """{} {{{}
}}
"""


replace_hook_body_fastcall = """
    /* In Py37 the replace method type moved to METH_FASTARGS. This means that
     * instead of args being passed as a tuple, they are passed as a C array
     * that contains PyObjects. We need to check whether there is the number of
     * args that we expect, and whether the arg we care about is not NULL.
     * Specifically, we want args[1] since it represents the "new" string in
     * the replacement.
     */
    PyObject *hook_args = pack_args_tuple(args, nargs);
    PyObject *result = {orig_func}{call};

    if (result == NULL || nargs < 2 || args[1] == NULL)
        goto cleanup_and_exit;

    if (result == self)
        goto cleanup_and_exit;

    call_string_propagator("propagate_{hook_name}", self, result, hook_args, NULL);

cleanup_and_exit:
    Py_XDECREF(hook_args);
    return result;
"""


split_hook_body_fastcall = """
    PyObject *result = {orig_func}(self, args, nargs, kwnames);
    PyObject *args_tuple = pack_args_tuple(args, nargs);
    PyObject *kwargs = pack_kwargs_dict(args, nargs, kwnames);

    if (result == NULL || PySequence_Length(result) == 1)
        goto cleanup_and_exit;

    call_string_propagator("propagate_{hook_name}", (PyObject *)self, result, args_tuple, kwargs);

cleanup_and_exit:
    Py_XDECREF(args_tuple);
    Py_XDECREF(kwargs);
    return result;
"""


join_hook_body = """
    PyObject *list = PySequence_List(args);

    /* Converting args to a list might legitimately raise an exception in some cases.
     * We need to be sure we don't suppress that exception to maintain original app
     * behavior. We've seen this before when args is a generator.
     */
    if (list == NULL) {{
        return NULL;
    }}

    /* In Py36+ we also hook an internal function that is called by this
     * function in order to propagate fstring formatting. We still want to have
     * a separate hook for join so that the events are reported differently.
     * This means that we need to go into scope when calling the original
     * function here so that we don't propagate twice.
     */
    enter_propagation_scope();
    PyObject *result = {orig_func}((PyObject *)self, list);
    exit_propagation_scope();

    PyObject *prop_args = PyTuple_Pack(1, list);

    if (prop_args == NULL || result == NULL)
        goto cleanup_and_exit;

    call_string_propagator("propagate_{hook_name}", (PyObject *)self, result, prop_args, NULL);

cleanup_and_exit:
    Py_XDECREF(list);
    Py_XDECREF(prop_args);
    return result;
"""


formatmap_hook_body = """
    PyObject *result = {orig_func}{call};

    if (result == NULL)
        return result;

    call_string_propagator("propagate_{hook_name}", self, result, NULL, args);

    return result;
"""


new_hook_body = """
    PyObject *result = {orig_func}{call};

    if (result == NULL)
        return result;

    call_string_propagator("propagate_{hook_name}", NULL, result, args, kwds);

    return result;
"""


init_hook_body = """
    int result = {orig_func}{call};

    if (result == -1)
        return result;

    /* Here we report self_obj=None and ret=self
       to maintain the illusion of casting */
    call_string_propagator("propagate_{hook_name}", NULL, self, args, kwds);

    return result;
"""


hook_macro_by_signature = {
    "unaryfunc": "HOOK_{}UNARYFUNC",
    "binaryfunc": "HOOK_{}BINARYFUNC",
    "ternaryfunc": "HOOK_{}TERNARYFUNC",
    fastcall_name: "HOOK_{}FASTCALL",
    fastcall_kwargs_name: "HOOK_TERNARY_FASTCALL",
}


special_methods = {
    "replace": (None, replace_hook_body_fastcall),
    "split": (None, split_hook_body_fastcall),
    "rsplit": (None, split_hook_body_fastcall),
    "join": (join_hook_body,) * 2,
    "formatmap": (formatmap_hook_body,) * 2,
    "new": (new_hook_body,) * 2,
    "init": (init_hook_body,) * 2,
}


def build_str_hook_macro(hook_name, signature):
    hook_definition = hook_macro_by_signature[signature]
    hook_definition = hook_definition.format("")
    hook_definition += "({});"
    return hook_definition.format(hook_name)


def build_stream_hook_macro(name, hook_name, signature):
    hook_definition = hook_macro_by_signature[signature]
    hook_definition = hook_definition.format("STREAM_")
    hook_definition += '({}, "{}");'
    return hook_definition.format(hook_name, name)


def build_special_hook(name, hook_name, signature):
    idx = 1 if signature in [fastcall_name, fastcall_kwargs_name] else 0

    hook_body = special_methods[name][idx]
    if hook_body is None:
        return build_str_hook_macro(hook_name, signature)

    hook_body = hook_body.format(
        orig_func="{}_orig".format(hook_name),
        hook_name=hook_name,
        call=calls_by_name[signature],
    )
    hook_funcname = "{}_new".format(hook_name)
    hook_signature = signatures_by_name[signature].format(hook_funcname)
    return hook_template.format(hook_signature, hook_body)


def create_hook(strtype, name, signature, stream_type):
    hook_name = "{}_{}".format(strtype, name)
    if name in special_methods:
        hook_definition = build_special_hook(name, hook_name, signature)
    elif stream_type:
        hook_definition = build_stream_hook_macro(name, hook_name, signature)
    else:
        hook_definition = build_str_hook_macro(hook_name, signature)

    return hook_definition, hook_name



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import ctypes
import sys
from contrast_vendor import structlog as logging
from contrast.agent import agent_lib

logger = logging.getLogger("contrast")


def initialize_semantic_analysis():
    if not agent_lib.modules.is_initialized():
        return

    agent_lib.modules.LIB_CONTRAST.does_file_path_bypass_security.argtypes = (
        ctypes.c_char_p,
        ctypes.c_char,
    )
    agent_lib.modules.LIB_CONTRAST.get_index_of_chained_command.argtypes = [
        ctypes.c_char_p
    ]
    agent_lib.modules.LIB_CONTRAST.does_command_contain_dangerous_path.argtypes = [
        ctypes.c_char_p
    ]

    agent_lib.modules.LIB_CONTRAST.does_file_path_bypass_security.restype = (
        ctypes.c_char
    )
    agent_lib.modules.LIB_CONTRAST.get_index_of_chained_command.restype = (
        ctypes.c_longlong
    )
    agent_lib.modules.LIB_CONTRAST.does_command_contain_dangerous_path.restype = (
        ctypes.c_char
    )


def get_index_of_chained_cmd(command: str) -> int:
    """
    @param command: str which contains the user input of the command to be executed
    @rtype int: -1 if there is no chained command or an int >=0 of the starting index
    of the chained command
    """
    if not command:
        return -1

    command = ctypes.c_char_p(bytes(command, "utf8", errors="ignore"))

    def is_valid_return(code):
        return code >= -1

    return agent_lib.call(
        agent_lib.modules.LIB_CONTRAST.get_index_of_chained_command,
        is_valid_return,
        command,
    )


def does_file_path_bypass_security(command: str, is_custom_code: object) -> int:
    """
    @param command: str which contains the user input of the command to be executed
    @param is_custom_code: object or none that indicates if code is coming from user code
    @rtype int: 1 if there is bypass in security or an int 0 if there is no security bypass
    """
    if not command:
        return 0

    command = ctypes.c_char_p(bytes(command, "utf8", errors="ignore"))

    is_custom_code_val = "0" if is_custom_code is None else "1"
    is_custom_code = ctypes.c_char(bytes(is_custom_code_val, "utf8", errors="ignore"))

    def is_valid_return(code):
        return 0 <= int.from_bytes(code, sys.byteorder) <= 1

    return int.from_bytes(
        agent_lib.call(
            agent_lib.modules.LIB_CONTRAST.does_file_path_bypass_security,
            is_valid_return,
            command,
            is_custom_code,
        ),
        sys.byteorder,
    )


def does_command_contain_dangerous_path(command):
    if not command:
        return 0

    command = ctypes.c_char_p(bytes(command, "utf8", errors="ignore"))

    def is_valid_return(code):
        return 0 <= int.from_bytes(code, sys.byteorder) <= 1

    return int.from_bytes(
        agent_lib.call(
            agent_lib.modules.LIB_CONTRAST.does_command_contain_dangerous_path,
            is_valid_return,
            command,
        ),
        sys.byteorder,
    )



import warnings

from contrast_vendor.webob.compat import (
    escape,
    string_types,
    text_,
    text_type,
    )

from contrast_vendor.webob.headers import _trans_key

def html_escape(s):
    """HTML-escape a string or object

    This converts any non-string objects passed into it to strings
    (actually, using ``unicode()``).  All values returned are
    non-unicode strings (using ``&#num;`` entities for all non-ASCII
    characters).

    None is treated specially, and returns the empty string.
    """
    if s is None:
        return ''
    __html__ = getattr(s, '__html__', None)
    if __html__ is not None and callable(__html__):
        return s.__html__()
    if not isinstance(s, string_types):
        __unicode__ = getattr(s, '__unicode__', None)
        if __unicode__ is not None and callable(__unicode__):
            s = s.__unicode__()
        else:
            s = str(s)
    s = escape(s, True)
    if isinstance(s, text_type):
        s = s.encode('ascii', 'xmlcharrefreplace')
    return text_(s)

def header_docstring(header, rfc_section):
    if header.isupper():
        header = _trans_key(header)
    major_section = rfc_section.split('.')[0]
    link = 'http://www.w3.org/Protocols/rfc2616/rfc2616-sec%s.html#sec%s' % (
        major_section, rfc_section)
    return "Gets and sets the ``%s`` header (`HTTP spec section %s <%s>`_)." % (
        header, rfc_section, link)


def warn_deprecation(text, version, stacklevel):
    # version specifies when to start raising exceptions instead of warnings
    if version in ('1.2', '1.3', '1.4', '1.5', '1.6', '1.7'):
        raise DeprecationWarning(text)
    else:
        cls = DeprecationWarning
    warnings.warn(text, cls, stacklevel=stacklevel + 1)

status_reasons = {
    # Status Codes
    # Informational
    100: 'Continue',
    101: 'Switching Protocols',
    102: 'Processing',

    # Successful
    200: 'OK',
    201: 'Created',
    202: 'Accepted',
    203: 'Non-Authoritative Information',
    204: 'No Content',
    205: 'Reset Content',
    206: 'Partial Content',
    207: 'Multi Status',
    226: 'IM Used',

    # Redirection
    300: 'Multiple Choices',
    301: 'Moved Permanently',
    302: 'Found',
    303: 'See Other',
    304: 'Not Modified',
    305: 'Use Proxy',
    307: 'Temporary Redirect',
    308: 'Permanent Redirect',

    # Client Error
    400: 'Bad Request',
    401: 'Unauthorized',
    402: 'Payment Required',
    403: 'Forbidden',
    404: 'Not Found',
    405: 'Method Not Allowed',
    406: 'Not Acceptable',
    407: 'Proxy Authentication Required',
    408: 'Request Timeout',
    409: 'Conflict',
    410: 'Gone',
    411: 'Length Required',
    412: 'Precondition Failed',
    413: 'Request Entity Too Large',
    414: 'Request URI Too Long',
    415: 'Unsupported Media Type',
    416: 'Requested Range Not Satisfiable',
    417: 'Expectation Failed',
    418: "I'm a teapot",
    422: 'Unprocessable Entity',
    423: 'Locked',
    424: 'Failed Dependency',
    426: 'Upgrade Required',
    428: 'Precondition Required',
    429: 'Too Many Requests',
    451: 'Unavailable for Legal Reasons',
    431: 'Request Header Fields Too Large',

    # Server Error
    500: 'Internal Server Error',
    501: 'Not Implemented',
    502: 'Bad Gateway',
    503: 'Service Unavailable',
    504: 'Gateway Timeout',
    505: 'HTTP Version Not Supported',
    507: 'Insufficient Storage',
    510: 'Not Extended',
    511: 'Network Authentication Required',
}

# generic class responses as per RFC2616
status_generic_reasons = {
    1: 'Continue',
    2: 'Success',
    3: 'Multiple Choices',
    4: 'Unknown Client Error',
    5: 'Unknown Server Error',
}

try:
    # py3.3+ have native comparison support
    from hmac import compare_digest
except ImportError: # pragma: nocover (Python 2.7.7 backported this)
    compare_digest = None

def strings_differ(string1, string2, compare_digest=compare_digest):
    """Check whether two strings differ while avoiding timing attacks.

    This function returns True if the given strings differ and False
    if they are equal.  It's careful not to leak information about *where*
    they differ as a result of its running time, which can be very important
    to avoid certain timing-related crypto attacks:

        http://seb.dbzteam.org/crypto/python-oauth-timing-hmac.pdf

    .. versionchanged:: 1.5
       Support :func:`hmac.compare_digest` if it is available (Python 2.7.7+
       and Python 3.3+).

    """
    len_eq = len(string1) == len(string2)
    if len_eq:
        invalid_bits = 0
        left = string1
    else:
        invalid_bits = 1
        left = string2
    right = string2

    if compare_digest is not None:
        invalid_bits += not compare_digest(left, right)
    else:
        for a, b in zip(left, right):
            invalid_bits += a != b
    return invalid_bits != 0




# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import contrast
from contrast.agent.protect.rule.base_rule import BaseRule
from contrast.api.user_input import InputType, UserInput

from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


class MethodTampering(BaseRule):
    RULE_NAME = "method-tampering"
    USER_INPUT_KEY = InputType.METHOD.name

    def postfilter(self):
        """
        At postfilter we generate activity if input analysis was found and depending on application response code.

        if response code is either 4xx or 5xx, application was not exploited (only probed) by an unexpected HTTP method.
        If response code is anything else, then an unexpected HTTP method successfully exploited the application.
        """
        logger.debug("PROTECT: Postfilter for %s", self.name)

        evaluations_for_rule = self.evaluations_for_rule()

        context = contrast.CS__CONTEXT_TRACKER.current()

        # do not remove; this case is not yet well-understood
        if (
            context is None
            or not hasattr(context, "response")
            or context.response is None
        ):
            logger.debug("WARNING: failed to get context in MethodTampering.postfilter")
            return

        response_code = context.response.status_code

        for evaluation in evaluations_for_rule:
            if str(response_code).startswith("4") or str(response_code).startswith("5"):
                attack = self.build_attack_without_match(
                    method=evaluation.value,
                    response_code=response_code,
                )
            else:
                attack = self.build_attack_with_match(
                    None,
                    method=evaluation.value,
                    response_code=response_code,
                )
            self._append_to_context(attack)

    def build_sample(self, evaluation, candidate_string, **kwargs):
        sample = self.build_base_sample(None)

        method = kwargs.get("method", "")

        sample.details["method"] = method
        sample.details["responseCode"] = kwargs.get("response_code", -1)
        sample.set_user_input(self.build_user_input(method))

        return sample

    def build_user_input(self, method):
        return UserInput(
            input_type=InputType.METHOD,
            key=self.USER_INPUT_KEY,
            value=method,
        )



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.assess.policy.propagators import SplatPropagator


class StarletteSafePathPropagator(SplatPropagator):
    """
    The purpose of this propagator is to SPLAT safe tags to the first element of the returned list from
    StaticFiles.lookup_path -> (full_path, os.stat(full_path))
    """

    def __init__(self, node, preshift, target):
        super(SplatPropagator, self).__init__(node, preshift, target)

        if isinstance(self.target, (tuple, list)):
            # First returned value of the list is the safely calculated path
            self.target = self.target[0]



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.assess.utils import get_properties
from contrast.agent.assess.policy.propagators.base_propagator import BasePropagator
from contrast.agent.policy.patch_manager import is_patched
from contrast.utils.decorators import fail_quietly

from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


class DBWritePropagator(BasePropagator):
    """
    Propagator to handle stored XSS. This propagator assumes the database
    column names are passed in via ALL_KWARGS.

    For each column name, patch the getter/setter property for the column name
    and create dynamic sources for these properties.
    """

    @property
    @fail_quietly(
        "Failed to determine if DB_WRITE needs propagation", return_value=False
    )
    def needs_propagation(self):
        """
        Here, the source is ALL_KWARGS. self.sources[0] is the kwarg dictionary.
        Each key is a column name, and each value is the value for this particular
        ORM instance (corresponding to a DB row) in that column.

        We need to check if any of the values are tracked strings. If so, we save
        all relevant information to be used in propagate().
        """

        self.new_column_sources = {}
        cls = self.preshift.obj.__class__
        source = self.sources[0]

        for col_name, value in source.items():
            if not value:
                continue

            if is_patched(getattr(cls, col_name, None)):
                continue

            col_value_properties = get_properties(value)
            if col_value_properties is None:
                continue

            self.new_column_sources[col_name] = col_value_properties

        return len(self.new_column_sources) > 0

    def track_target(self):
        """
        This propagator does not track the target because the target is
        an instance of an ORM model, such as django's Model class. If it
        were a string instead, we would track it.
        """
        pass

    def add_tags_and_properties(self, ret):
        """
        Because the target is an instance of an ORM model,
        such as django's Model class, and not a string, we do this
        work inside self.propagate(). Later on we could refactor
        to move the work here.
        """
        pass

    def propagate(self):
        from contrast.agent.policy.applicator import apply_patch_to_dynamic_property

        cls = self.preshift.obj.__class__

        for col_name, col_value_properties in self.new_column_sources.items():
            self.apply_tags(self.node, col_value_properties.origin)

            tags = [tag for tag in col_value_properties.tags]
            logger.debug("DB_WRITE adding column as dynamic source: %s", col_name)
            apply_patch_to_dynamic_property(cls, col_name, tags)

            col_value_properties.build_event(
                self.node,
                col_value_properties.origin,
                self.preshift.obj,
                self.target,
                self.preshift.args,
                self.preshift.kwargs,
                [],
                0,
                None,
            )



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import os
import hashlib

CONTRAST_AGENT_DIST = "contrast-agent"

# Both of these metadata files contain a file list of what is installed under the top level dirs
RECORD = "RECORD"
SOURCES = "SOURCES.txt"

NAMESPACE_PACKAGE = "namespace_packages.txt"
TOP_LEVEL_TXT = "top_level.txt"

PY_SUFFIX = ".py"
SO_SUFFIX = ".so"

SITE_PACKAGES_DIR = f"{os.sep}site-packages{os.sep}"
DIST_PACKAGES_DIR = f"{os.sep}dist-packages{os.sep}"

from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


def normalize_file_name(file_path):
    """
    This function converts a file ending in .pyc to .py. The reason for this
    is due to how screener is configured to verify a file was reported (only supports
    exact match not a regex)
    @param file_path: full path to a python file ending in .pyc or .py
    @return: file_path ending in .py
    """
    file_to_report = file_path.rpartition(SITE_PACKAGES_DIR)

    if not file_to_report[1]:
        file_to_report = file_path.rpartition(DIST_PACKAGES_DIR)
        if not file_to_report[1]:
            return None

    normalized_file_name = file_to_report[2]
    if normalized_file_name.endswith(".pyc"):
        normalized_file_name = normalized_file_name[: len(normalized_file_name) - 1]

    return normalized_file_name


def _parse_top_level_dirs_from_manifest_file(dist, namespace, metadata_filename):
    """
    @param dist: Distribution object used to check to see what
    metadata files exist for us to parse
    @type dist: pkg_resources.DistInfoDistribution
    @param namespace: The name of the namespace to search
    @type namespace: string
    @param metadata_filename: Name of metadata file to parse
    @type metadata_filename: string
    """
    top_level_dirs = set()

    if dist.has_metadata(metadata_filename):
        for line in dist.get_metadata_lines(metadata_filename):
            if line.startswith(namespace + os.sep):
                dirs = line.split(os.sep)
                if len(dirs) > 1:
                    top_level_dirs.add(dirs[1])

    return top_level_dirs


def get_top_level_directories_namespace_pkg(dist, namespace):
    """
    @param dist: Distribution object used to check to see what
    metadata files exist for us to parse
    @type dist: pkg_resources.DistInfoDistribution
    @param namespace: The name of the namespace to search
    @type namespace: string
    @return: The top level importable packages/modules under the namespace
    @rtype: string
    """
    top_level_dirs = set()
    manifest_files = (RECORD, SOURCES)

    if not dist:
        return top_level_dirs

    for manifest in manifest_files:
        top_level_dirs = _parse_top_level_dirs_from_manifest_file(
            dist, namespace, manifest
        )
        if top_level_dirs:
            break

    return top_level_dirs


def get_top_level_directories(dist):
    """
    Parse the top_level.txt or RECORD file for a dist to get the dist's
    top level directories.

    :param dist: instance of pkg_resources.DistInfoDistribution
    :return: list of strs representing names of top level dist directories
    """
    top_level_dirs = []

    if dist.has_metadata(TOP_LEVEL_TXT):
        top_level_dirs = list(dist.get_metadata_lines(TOP_LEVEL_TXT))
    elif dist.has_metadata(RECORD):
        top_level_dirs = _parse_record_lines(dist.get_metadata_lines(RECORD))
    else:
        logger.debug("Cannot find top level dirs for %s", dist)

    return top_level_dirs


def _parse_record_lines(record_data_generator):
    """
    For lines in a RECORD file, parse out the name(s) of the top level directories.

    :param record_data_generator: generator containing lines in RECORD file
    :return: list of strs representing names of top level dist directories
    """
    top_level_dirs = []
    for full_name in record_data_generator:
        name = full_name.split(",")[0]
        if os.sep in name and name.split(os.sep)[1] == "__init__.py":
            top_level_dirs.append(name.split(os.sep)[0])
        elif name.endswith(SO_SUFFIX):
            # Handle special case where an .so file is a top level importable module.
            # This library is loaded by importing the name of the file excluding
            # the platform specific name convention of the file
            # i.e remove .cpython-PY_VERSION-platform.so from module.cpython-PY_VERSION-platform.so
            top_level_dirs.append(name.split(".")[0])

    return top_level_dirs


def get_file_from_module(module):
    if hasattr(module, "__file__") and module.__file__:
        return os.path.realpath(module.__file__)

    return None


def get_url_from_dist(dist):
    """
    Gets the library url from either PKG-INFO or METADATA files
    :param dist: the distribution package where data is being parsed
    :return: the url of the package or "" if one wasn't discovered
    """
    home_page = "Home-page: "
    url = ""

    metadata = list(dist.get_metadata_lines(dist.PKG_INFO))
    for line in metadata:
        if line.startswith(home_page):
            url = line.split(home_page)[1]
            break
    else:
        logger.debug("Cannot find url for %s", dist)

    return url


def get_data(dist):
    """
    Given a dist, pulls name, version, manifest, and url out of the metadata
    :param dist: the distribution package whose package info is being retrieved
    :return: the package info from the metadata
    """
    version = dist.version
    manifest = dist.get_metadata(dist.PKG_INFO)
    url = get_url_from_dist(dist)
    return version, manifest, str(url)


def get_hash(name, version):
    """
    DO NOT ALTER OR REMOVE
    """
    to_hash = name + " " + version

    return hashlib.sha1(to_hash.encode("utf-8")).hexdigest()



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
"""
Implementation of patches for the `re` module.

Other methods of the `re` module such as `split` and `escape` can be patched directly,
so they are implemented in policy.
"""
import functools
from contrast_vendor import structlog as logging
import sys

from contrast_vendor.wrapt import register_post_import_hook

import contrast
from contrast.agent import scope
from contrast.agent.policy import patch_manager, registry
from contrast.agent.assess.apply_trigger import cs__apply_trigger
from contrast.agent.assess.policy.propagators import regex_propagator
from contrast.patches.utils import analyze_policy, get_propagation_node
from contrast.utils.decorators import fail_quietly
from contrast.utils.patch_utils import (
    build_and_apply_patch,
    pack_self,
    wrap_and_watermark,
)

logger = logging.getLogger("contrast")

PATTERN_CLASS = "re.Pattern"
MATCH_CLASS = "re.Match"


def build_group_hook(original_func, patch_policy, propagator):
    del patch_policy

    def group_hook(wrapped, instance, args, kwargs):
        context = contrast.CS__CONTEXT_TRACKER.current()
        result = wrapped(*args, **kwargs)
        if (
            context is not None
            and context.propagate_assess
            and not context.stop_propagation
        ):
            context.propagated()
            with scope.propagation_scope():
                propagator(instance, result, *args)
        return result

    return wrap_and_watermark(original_func, group_hook)


def wrap_repl(repl):
    repl_results = []

    def new_repl(match):
        result = repl(match)
        repl_results.append(result)
        return result

    functools.update_wrapper(new_repl, repl)
    return new_repl, repl_results


@fail_quietly("Failed to propagate sub(n)")
def _analyze_sub(node, retval, repl_results, args, kwargs, new_args):
    if node.method_name == "subn":
        result, count = retval
    else:
        # Account for the fact that count could either be positional or kwarg
        count = args[3] if len(args) == 4 else kwargs.pop("count", 0)
        result = retval

    # Omit count (and flags) if they are part of posargs since they are being passed
    # explicitly to our propagator
    new_args = new_args[:3]
    new_kwargs = dict(count=count)
    if not node.instance_method:
        # Account for the fact that flags could either be positional or kwarg
        new_kwargs["flags"] = args[4] if len(args) > 4 else kwargs.get("flags", 0)

    with scope.propagation_scope():
        regex_propagator.propagate_sub(
            node, result, repl_results, *new_args[:3], **new_kwargs
        )


@fail_quietly("Failed to analyze redos trigger")
def _trigger_redos(name, result, args, kwargs):
    context = contrast.CS__CONTEXT_TRACKER.current()
    if context is None:
        return

    rule = registry.get_triggers_by_rule("redos")

    if not scope.in_trigger_scope() and not scope.in_scope():
        if rule.disabled:
            # Given how often these patches are called within apps, it's well worth
            # returning early here even though this logic exists further down the stack.
            # However, we MUST call this here after the `in_scope` check because
            # `rule.disabled` calls on Settings code that recursively calls on
            # other patches.
            return

        trigger_node = _get_redos_trigger_node(name, rule)

        # Both trigger and contrast scope are needed here.
        with scope.trigger_scope(), scope.contrast_scope():
            # we cannot use trigger_policy.apply here due to the instance_method
            # logic it uses to remove self_obj from args. For the redos action,
            # specifically for re.Pattern.method nodes, we NEED the self_obj
            # to be in the args at this time.
            source = trigger_node.get_matching_sources(None, result, args, kwargs)[0]
            cs__apply_trigger(
                context,
                rule,
                trigger_node,
                source,
                None,
                result,
                None,
                args,
                kwargs,
            )


def _get_redos_trigger_node(trigger_name, rule):
    trigger_loc = "re"
    if "Pattern" in trigger_name:
        trigger_loc = PATTERN_CLASS

    func_name = trigger_name.split(".")[-1]
    trigger_nodes = rule.find_trigger_nodes(trigger_loc, func_name)

    return trigger_nodes[0]


def build_sub_hook(original_func, policy_node):
    policy_node = get_propagation_node(policy_node)

    def sub_hook(wrapped, instance, args, kwargs):
        """
        Hook for re.sub and re.subn used for propagation in assess

        The following explains why we can't simply patch these methods using
        policy.

        It is possible for the repl argument to be a callable. In this case, the
        callable is passed a Match object, and it returns the string to be used for
        the replacement. In order to correctly propagate the substitution
        operation, we need to keep track of the results of calling the replacement
        function.

        It might seem like we should just call the replacement function again
        during our propagation action. But this is not practicable for several
        reasons:

          1. We're in scope at the time, so any propagation that needs to occur
             within the replacement callable itself will be missed.
          2. Related to above, but methods of Match do not return the same object
             even when called multiple times with the same arguments, so we would
             not be tracking the strings that actually get used in the substitution
             result.
          3. There's no guarantee that the replacement function does not cause any
             side effects or rely on any state in application code. We definitely
             don't want to mess around with this.

        The solution is to wrap the replacement callable with our own function that
        records the results of each call. We then pass our wrapped callable to the
        original function, and we pass the accumulated results to the propagator.
        This has the additional benefit of allowing us to wrap the match object
        that is passed to the repl function with our proxied object so that we
        propagate any calls that are made within this function if necessary.
        """
        # Get the non-propagation case out of the way here
        context = contrast.CS__CONTEXT_TRACKER.current()
        if context is None or not context.propagate_assess or context.stop_propagation:
            return wrapped(*args, **kwargs)
        context.propagated()

        packed_args = pack_self(instance, args)

        try:
            repl = packed_args[1]
            new_repl, repl_results = wrap_repl(repl) if callable(repl) else (repl, None)
            new_args = tuple(packed_args[:1]) + (new_repl,) + tuple(packed_args[2:])
        except Exception:
            # This indicates that the original caller passed garbage, so call the original
            # function and let the error propagate back up to where they can clean up their
            # own mess.
            return wrapped(*args, **kwargs)

        # Account for the fact that the call to wrapped never takes a self
        call_args_idx = 1 if instance is not None else 0
        retval = wrapped(*new_args[call_args_idx:], **kwargs)

        _trigger_redos(policy_node.name, retval, packed_args, kwargs)
        _analyze_sub(policy_node, retval, repl_results, packed_args, kwargs, new_args)

        return retval

    return wrap_and_watermark(original_func, sub_hook)


def build_re_assess_hook(original_func, patch_policy):
    """Builds patch that analyzes for both redos trigger and regex propagation"""

    def assess_hook(wrapped, instance, args, kwargs):
        result = wrapped(*args, **kwargs)
        args = pack_self(instance, args)
        _trigger_redos(patch_policy.name, result, args, kwargs)
        analyze_policy(patch_policy.name, result, args, kwargs)
        return result

    return wrap_and_watermark(original_func, assess_hook)


def build_re_pattern_redos_hook(original_func, _):
    """
    Builds patch for re.Pattern methods that performs analysis for redos trigger only
    """

    method_name = original_func.__name__
    full_name = "re.Pattern." + method_name

    def redos_hook(wrapped, instance, args, kwargs):
        retval = wrapped(*args, **kwargs)
        _trigger_redos(full_name, retval, pack_self(instance, args), kwargs)
        return retval

    return wrap_and_watermark(original_func, redos_hook)


def patch_re(re_module):
    # The re.Pattern and re.Match classes are not directly accessible in all versions
    # of Python, so we do this somewhat hacky workaround to get a reference to them.
    pattern_cls = re_module.compile("").__class__
    match_cls = re_module.match("", "").__class__

    build_and_apply_patch(re_module, "sub", build_sub_hook)
    build_and_apply_patch(re_module, "subn", build_sub_hook)

    build_and_apply_patch(pattern_cls, "sub", build_sub_hook, owner_name=PATTERN_CLASS)
    build_and_apply_patch(pattern_cls, "subn", build_sub_hook, owner_name=PATTERN_CLASS)
    build_and_apply_patch(
        pattern_cls,
        "split",
        build_re_assess_hook,
        owner_name=PATTERN_CLASS,
    )
    build_and_apply_patch(
        pattern_cls,
        "findall",
        build_re_assess_hook,
        owner_name=PATTERN_CLASS,
    )

    build_and_apply_patch(
        pattern_cls,
        "match",
        build_re_pattern_redos_hook,
        owner_name=PATTERN_CLASS,
    )
    build_and_apply_patch(
        pattern_cls,
        "search",
        build_re_pattern_redos_hook,
        owner_name=PATTERN_CLASS,
    )
    build_and_apply_patch(
        pattern_cls, "finditer", build_re_pattern_redos_hook, owner_name=PATTERN_CLASS
    )

    build_and_apply_patch(
        pattern_cls,
        "fullmatch",
        build_re_pattern_redos_hook,
        owner_name=PATTERN_CLASS,
    )

    build_and_apply_patch(
        match_cls,
        "group",
        build_group_hook,
        (regex_propagator.propagate_group,),
        owner_name=MATCH_CLASS,
    )
    build_and_apply_patch(
        match_cls,
        "groups",
        build_group_hook,
        (regex_propagator.propagate_groups,),
        owner_name=MATCH_CLASS,
    )
    build_and_apply_patch(
        match_cls,
        "groupdict",
        build_group_hook,
        (regex_propagator.propagate_groupdict,),
        owner_name=MATCH_CLASS,
    )


def register_patches():
    register_post_import_hook(patch_re, "re")


def reverse_patches():
    re_module = sys.modules.get("re")
    if not re_module:
        return

    pattern_cls = re_module.compile("").__class__
    match_cls = re_module.match("", "").__class__

    patch_manager.reverse_patches_by_owner(re_module)
    patch_manager.reverse_patches_by_owner(pattern_cls)
    patch_manager.reverse_patches_by_owner(match_cls)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
"""
Builds/defines propagation functions to be called from C extension hooks

NOTE: This module will need to be loaded before the extension executes
"""
import sys

from contrast_vendor import structlog as logging

import contrast
from contrast.agent.assess.policy.preshift import Preshift
from contrast.agent.assess.policy.propagation_policy import track_copy_without_new_event

# We call these from the C extension - they're import aliases
from contrast.agent.assess.policy.propagation_policy import (  # pylint: disable=unused-import
    create_stream_source_event,
    propagate_stream,
)
from contrast.agent.assess.policy.propagators import BasePropagator, PROPAGATOR_ACTIONS
from contrast.agent.policy import registry


HOOK_TYPE_NAMES = ["unicode", "bytes", "bytearray"]


logger = logging.getLogger("contrast")
module = sys.modules[__name__]


def build_generic_strtype_propagator(policy_node):
    """
    Build propagator function used for most methods of unicode and bytes types

    By not accounting for bytearray propagation in this function we are able to
    optimize for the most frequent propagation events.

    We also optimize by avoiding handling for a few unicode/bytes special case methods
    in this function. Those special cases are handled by other builders below.

    While this leads to some code duplication, it allows us to optimize some of the
    most intensively called code in the agent.
    """
    propagator_class = PROPAGATOR_ACTIONS.get(policy_node.action, BasePropagator)

    def propagate_string(target, self_obj, ret, args, kwargs):
        # Short-circuit that applies to all string types except bytearray
        if ret is self_obj:
            return

        context = contrast.CS__CONTEXT_TRACKER.current()
        if context is None or context.stop_propagation:
            return

        preshift = Preshift(self_obj, args or [], kwargs or {})

        propagator = propagator_class(policy_node, preshift, target)
        if not propagator.needs_propagation:
            return

        propagator.track_and_propagate(ret)

        logger.debug(
            "Propagator %s found: propagated to %s", policy_node.name, id(target)
        )

        context.propagated()

    return propagate_string


def build_bytearray_propagator(policy_node):
    """
    Build propagator function used for all methods of bytearray instances

    Copies of bytearrays return a new object, so it is necessary for us to track the
    new string, even if it hasn't changed.

    We could break the concat special case into a separate propagator function, but I'm
    not really sure it's worth it for bytearray since bytearay events are expected to
    be pretty rare. We can change it if it ever becomes a performance issue.

    Keeping this case separate from the generic builder allows us to avoid slowing down
    the propagation of the more frequently called unicode/bytes propagators, even if it
    leads to some duplicated code.
    """
    propagator_class = PROPAGATOR_ACTIONS.get(policy_node.action, BasePropagator)
    method_name = policy_node.method_name

    def propagate_bytearray(target, self_obj, ret, args, kwargs):
        context = contrast.CS__CONTEXT_TRACKER.current()
        if context is None or context.stop_propagation:
            return

        if ret == self_obj and method_name != "concat":
            track_copy_without_new_event(target, self_obj)
            return

        preshift = Preshift(self_obj, args or [], kwargs or {})

        propagator = propagator_class(policy_node, preshift, target)
        if not propagator.needs_propagation:
            return

        propagator.track_and_propagate(ret)

        logger.debug(
            "Propagator %s found: propagated to %s", policy_node.name, id(target)
        )

        context.propagated()

    return propagate_bytearray


def build_translate_propagator(policy_node):
    """
    Build propagator function used in unicode.translate hook

    Py27 unicode objects and Py3x str objects returned by str.translate() are always
    new objects, so in order to keep our behavior consistent we do not create a new
    event in the case where new == original.

    Having a separate builder here allows us to optimize for the translate special case
    without imposing a blanket performance overhead on all other unicode/bytes methods.
    """
    propagator_class = PROPAGATOR_ACTIONS.get(policy_node.action, BasePropagator)

    def propagate_translate(target, self_obj, ret, args, kwargs):
        context = contrast.CS__CONTEXT_TRACKER.current()
        if context is None or context.stop_propagation:
            return

        if ret == self_obj:
            track_copy_without_new_event(target, self_obj)
            return

        preshift = Preshift(self_obj, args or [], kwargs or {})

        propagator = propagator_class(policy_node, preshift, target)
        if not propagator.needs_propagation:
            return

        propagator.track_and_propagate(ret)

        logger.debug(
            "Propagator %s found: propagated to %s", policy_node.name, id(target)
        )

        context.propagated()

    return propagate_translate


def build_cformat_propagator(policy_node):
    """
    Build propagator function used in the unicode/bytes cformat hook

    Account for the fact that None is a valid argument for cformat, and it also may be
    a single object. The preshift requires an argument arrary.  Eventually this should
    maybe just move into the C hook for cformat, although it is a bit easier to
    implement (and debug) where it is here.

    Having a separate builder here allows us to optimize for the translate special case
    without imposing a blanket performance overhead on all other unicode/bytes methods.
    """
    propagator_class = PROPAGATOR_ACTIONS.get(policy_node.action, BasePropagator)

    def propagate_cformat(target, self_obj, ret, args, kwargs):
        # Short-circuit that applies to all string types except bytearray
        if ret is self_obj:
            return

        context = contrast.CS__CONTEXT_TRACKER.current()
        if context is None or context.stop_propagation:
            return

        if not isinstance(args, tuple):
            args = (args,)

        preshift = Preshift(self_obj, args or [], kwargs or {})

        propagator = propagator_class(policy_node, preshift, target)
        if not propagator.needs_propagation:
            return

        propagator.track_and_propagate(ret)

        logger.debug(
            "Propagator %s found: propagated to %s", policy_node.name, id(target)
        )

        context.propagated()

    return propagate_cformat


def build_generic_cast_propagator(policy_node):
    """
    Build propagator function used for cast propagation for unicode/bytes

    Having a separate builder here allows us to optimize for the translate special case
    without imposing a blanket performance overhead on all other unicode/bytes methods.
    """
    propagator_class = PROPAGATOR_ACTIONS.get(policy_node.action, BasePropagator)

    def propagate_cast(target, self_obj, ret, args, kwargs):
        if ret is (args and args[0]):
            return

        context = contrast.CS__CONTEXT_TRACKER.current()
        if context is None or context.stop_propagation:
            return

        preshift = Preshift(self_obj, args or [], kwargs or {})

        propagator = propagator_class(policy_node, preshift, target)
        if not propagator.needs_propagation:
            return

        propagator.track_and_propagate(ret)

        logger.debug(
            "Propagator %s found: propagated to %s", policy_node.name, id(target)
        )

        context.propagated()

    return propagate_cast


def build_bytearray_cast_propagator(policy_node):
    """
    Build propagator function used for cast propagation for bytearray

    Having a separate builder here allows us to optimize for the translate special case
    without imposing a blanket performance overhead on all other unicode/bytes methods.
    """
    propagator_class = PROPAGATOR_ACTIONS.get(policy_node.action, BasePropagator)

    def propagate_cast(target, self_obj, ret, args, kwargs):
        source = args and args[0]
        if (
            ret == source
            and isinstance(target, bytearray)
            and isinstance(source, bytearray)
        ):
            track_copy_without_new_event(target, source)
            return

        context = contrast.CS__CONTEXT_TRACKER.current()
        if context is None or context.stop_propagation:
            return

        preshift = Preshift(self_obj, args or [], kwargs or {})

        propagator = propagator_class(policy_node, preshift, target)
        if not propagator.needs_propagation:
            return

        propagator.track_and_propagate(ret)

        logger.debug(
            "Propagator %s found: propagated to %s", policy_node.name, id(target)
        )

        context.propagated()

    return propagate_cast


def _create_propagator_function(strtype, location, propagation_node, builder):
    propagator_name = f"propagate_{strtype}_{location.method_name.lower()}"
    propagator = builder(propagation_node)
    propagator.__name__ = propagator_name
    setattr(module, propagator_name, propagator)


PROPAGATION_SPECIAL_CASES = {
    ("unicode", "translate"): build_translate_propagator,
    ("unicode", "cformat"): build_cformat_propagator,
    ("bytes", "cformat"): build_cformat_propagator,
    ("unicode", "CAST"): build_generic_cast_propagator,
    ("bytes", "CAST"): build_generic_cast_propagator,
    ("bytearray", "CAST"): build_bytearray_cast_propagator,
}


def build_string_propagator_functions():
    """
    Builds propagator functions that will be called from C extension hooks

    We don't particularly care if we create a few functions that don't correspond to
    any real hooks (e.g. bytearray.format). This is because if the hook doesn't exist,
    then the function will never be called.
    """
    for location in registry.get_string_method_nodes():
        prop_node = location.propagator_nodes[0]
        for strtype in HOOK_TYPE_NAMES:
            # Check whether there's a special case for this hook location
            propagator = PROPAGATION_SPECIAL_CASES.get(
                (strtype, location.method_name),
                (
                    # General case for either bytearray or other strtype
                    build_bytearray_propagator
                    if strtype == "bytearray"
                    else build_generic_strtype_propagator
                ),
            )
            _create_propagator_function(strtype, location, prop_node, propagator)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import hashlib
import traceback
import inspect
from contrast.agent.validator import Validator
from contrast.utils import get_installed_distributions
from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


class _Exception(Validator):
    VALIDATIONS = dict(
        type=dict(required=True, range=(1, 256)),
        module=dict(required=False, range=(1, 256)),
        value=dict(required=False, range=(0, 256)),
        stackframes=dict(required=True, range=(0, 128)),
    )

    def __init__(self, error: Exception):
        self.type = error.__class__.__name__
        # module would be location where this error is defined, not where raised.
        self.module = self._determine_module(error)
        self.value = str(error)
        self.stackframes = self._build_stackframes(error)

        self.validate()

    def _determine_module(self, error):
        """
        Find the str representing the path for where an error / exception is defined.
        """
        module = ""

        try:
            module = inspect.getfile(error.__class__)
        except TypeError as ex:
            if "is a built-in class" in str(ex):
                module = "builtin"

        return redact(module)

    def _build_stackframes(self, error):
        frames = []
        stacks = traceback.extract_tb(error.__traceback__)
        for stack in stacks:
            frames.append(
                _StackFrame(
                    # type is more appropriate for other languages so for now just
                    # make it the function name
                    stack_line=stack.line,
                    module=stack.filename,
                    function=stack.name,
                    lineno=stack.lineno,
                )
            )
        return frames

    def __repr__(self):
        return f"{self.__class__.__name__}: {self.to_json()}"

    def to_json(self):
        return dict(
            type=self.type,
            module=self.module,
            value=self.value,
            stackFrames=[ex.to_json() for ex in self.stackframes],
        )


class _StackFrame(Validator):
    VALIDATIONS = dict(
        line=dict(required=True, range=(1, 256)),
        lineno=dict(required=False),
        module=dict(required=False, range=(1, 256)),
        function=dict(required=True, range=(1, 256)),
        in_contrast=dict(required=False, default=False),
    )

    def __init__(self, stack_line, module, function, lineno):
        self.line = stack_line
        self.in_contrast = "/contrast" in module
        # Don't redact if we're in contrast code
        self.module = module if self.in_contrast else redact(module)
        self.function = function
        self.lineno = lineno

        self.validate()

    def __repr__(self):
        return f"{self.__class__.__name__}: {self.to_json()}"

    def to_json(self):
        return dict(
            # type is meant to represent the more primitive types of functional langs,
            # for python we will make this the line
            type=self.line,
            module=self.module,
            function=self.function,
            fileLineNumber=self.lineno,
            inContrast=self.in_contrast,
        )


def redact(text):
    """
    Determine if to redact `text` by converting it to sha256 if
    `text` does not contain an accepted name.

    When in doubt, redact. We want to prevent reporting private customer data.

    :param text:
    :return: `text` or sha256 version of text
    """

    if not text:
        return ""

    text_lower = text.lower()

    accepted_third_party = ["contrast", "builtin"] + get_installed_distributions()

    for accepted in accepted_third_party:
        name = (
            accepted.lower()
            if isinstance(accepted, str)
            else accepted.project_name.lower()
        )

        # length check is a fail safe to not do a too-short str comparison.
        if len(name) > 2 and name in text_lower:
            return text

    return hashlib.sha256(text.encode()).hexdigest()[:7]



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import re

from contrast.agent.protect.rule.base_rule import BaseRule
from contrast.agent.protect.rule.xxe.entity_wrapper import EntityWrapper
from contrast.api.user_input import UserInput, InputType
from contrast.api.sample import Sample
from contrast.utils.stack_trace_utils import build_and_clean_stack


class Xxe(BaseRule):
    """
    XXE Protection rule
    """

    RULE_NAME = "xxe"
    INPUT_NAME = "XML Prolog"

    EXTERNAL_ENTITY_PATTERN = re.compile(
        r"<!ENTITY\s+[a-zA-Z0-f]+\s+(?:SYSTEM|PUBLIC)\s+(.*?)>"
    )

    def __init__(self):
        super().__init__()

        self.prolog_xml = None

    def is_prefilter(self):
        return False

    def is_postfilter(self):
        return False

    def find_attack(self, candidate_string=None, **kwargs):
        if self.protect_excluded_by_code():
            return None

        last_idx = 0
        declared_entities = []
        entities_resolved = []

        for match in self.EXTERNAL_ENTITY_PATTERN.finditer(candidate_string):
            last_idx = match.end(0)

            entity_wrapper = EntityWrapper(match.group())
            if not entity_wrapper.is_external_entity():
                continue

            declared_entities.extend([self._build_match(match.group(), last_idx)])
            entities_resolved.extend([self._build_wrapper(entity_wrapper)])

        self.prolog_xml = candidate_string[:last_idx]

        attack = self.build_attack_with_match(
            candidate_string,
            declared_entities=declared_entities,
            entities_resolved=entities_resolved,
        )
        return attack

    def _build_match(self, match_string, end_idx):
        return {"start": end_idx - len(match_string), "end": end_idx}

    def _build_wrapper(self, entity):
        return {"systemId": entity.system_id, "publicId": entity.public_id}

    def build_sample(self, evaluation, url, **kwargs):
        sample = self.build_base_sample(evaluation)
        declared_entities = kwargs["declared_entities"]
        if declared_entities:
            sample.details["declaredEntities"] = declared_entities

        entities_resolved = kwargs["entities_resolved"]
        if entities_resolved:
            sample.details["entitiesResolved"] = entities_resolved

        sample.details["xml"] = self.prolog_xml
        return sample

    def build_base_sample(self, evaluation, **args):
        sample = Sample()
        sample.set_user_input(self.build_user_input(evaluation))
        sample.set_stack(build_and_clean_stack())
        return sample

    def build_user_input(self, _):
        return UserInput(
            input_type=InputType.UNKNOWN,
            key=self.INPUT_NAME,
            value=self.prolog_xml,
        )

    def infilter_kwargs(self, user_input, patch_policy):
        return dict(framework=patch_policy.method_name)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
test_environ = {
    "wsgi.version": (1, 0),
    "wsgi.url_scheme": "http",
    "wsgi.multithread": True,
    "wsgi.multiprocess": False,
    "wsgi.run_once": False,
    "SERVER_SOFTWARE": "Werkzeug/2.0.2",
    "REQUEST_METHOD": "GET",
    "SCRIPT_NAME": "",
    "PATH_INFO": "/",
    "QUERY_STRING": "",
    "REQUEST_URI": "/",
    "RAW_URI": "/",
    "REMOTE_ADDR": "127.0.0.1",
    "REMOTE_PORT": 61142,
    "SERVER_NAME": "127.0.0.1",
    "SERVER_PORT": "8000",
    "SERVER_PROTOCOL": "HTTP/1.1",
    "HTTP_HOST": "localhost:8000",
    "HTTP_CONNECTION": "keep-alive",
    "HTTP_SEC_CH_UA": "",
    "HTTP_SEC_CH_UA_MOBILE": "?0",
    "HTTP_SEC_CH_UA_PLATFORM": '"macOS"',
    "HTTP_UPGRADE_INSECURE_REQUESTS": "1",
    "HTTP_USER_AGENT": "",
    "HTTP_ACCEPT": "text/html",
    "HTTP_SEC_FETCH_SITE": "none",
    "HTTP_SEC_FETCH_MODE": "navigate",
    "HTTP_SEC_FETCH_USER": "?1",
    "HTTP_SEC_FETCH_DEST": "document",
    "HTTP_ACCEPT_ENCODING": "gzip, deflate, br",
    "HTTP_ACCEPT_LANGUAGE": "en-US,en;q=0.9",
}



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
"""
Rule applicator for SQLi for dbapi2-compliant modules
"""
import contrast
from contrast.agent import scope
from contrast.agent.protect.rule.sqli_rule import SqlInjection
from contrast.agent.policy import registry
from contrast.agent.assess.policy.analysis import skip_analysis
from contrast.agent.assess.policy import trigger_policy
from contrast.agent.settings import Settings
from contrast.utils.decorators import fail_quietly


@fail_quietly("Error running SQLi assess rule")
def assess_rule(context, adapter_name, method, result, args, kwargs):
    """
    Apply assess SQLi rule

    @param context: Current RequestContext instance
    @param adapter_name: String representing the adapter module (e.g. "sqlite3")
    @param method: String representing the database method (e.g. "execute")
    @param result: Object representing result of database action
    @param args: The args tuple passed to the original database function
    @param kwargs: The kwargs dict passed to the original database function
    """
    if scope.in_trigger_scope():
        return

    rule = registry.get_triggers_by_rule("sql-injection")
    trigger_nodes = rule.find_trigger_nodes(f"{adapter_name}.Cursor", method)

    trigger_policy.apply(rule, trigger_nodes, result, args, kwargs)


def apply_rule(adapter_name, orig_func, args, kwargs):
    """
    Common API for applying SQLi rule (applies both protect and assess rules)

    Applies the assess rule if assess is enabled. If protect is enabled,
    applies the protect rule. Important caveats:
      - neither rule will be applied if there is not an active request context
      - the assess rule will not be applied if we are already in scope

    The protect rule *must* be applied prior to calling the original function.
    This is the only way that we can raise a SecurityException if the rule is
    in BLOCK mode.

    The assess rule *must* be applied even if the original function results in
    an exception. Otherwise we won't detect vulnerable dataflows that may have
    just happened to result in an exception under testing.

    @param adapter_name: String representing the adapter (e.g. "sqlite3")
    @param orig_func: Original function (i.e. the one replaced by the patch)
    @param args: The args passed to the original database function
    @param kwargs: The kwargs passed to the original database function

    @return: Returns the object that is returned by calling orig_func
    """
    context = contrast.CS__CONTEXT_TRACKER.current()
    result = None

    if context is not None:
        _protect_rule(adapter_name, orig_func.__name__, args, kwargs)

    try:
        result = orig_func(*args[1:], **kwargs)
    finally:
        if not skip_analysis(context):
            assess_rule(context, adapter_name, orig_func.__name__, result, args, kwargs)

    return result


@fail_quietly("Error running SQLi protect rule")
def _protect_rule(adapter_name, method, args, kwargs):
    """
    Run protect SQLi analysis if it's enabled.
    """
    if not Settings().is_protect_enabled() or len(args) <= 1:
        return

    # args[0] is self, args[1] is the query
    sql = args[1]

    # create fake policy node and patch loc policy only to fit the general api.
    from contrast.agent.policy.patch_location_policy import PatchLocationPolicy
    from contrast.agent.policy.trigger_node import TriggerNode

    trigger_node = TriggerNode(adapter_name, "unused", True, method, "ARG_0")
    patch_policy = PatchLocationPolicy(trigger_node)

    rule = Settings().protect_rules.get(SqlInjection.RULE_NAME)
    if not rule or not rule.enabled:
        return

    rule.protect(patch_policy, sql, args, kwargs)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.assess.rules.dataflow_rule import DataflowRule
from contrast.agent.policy.registry import register_trigger_rule

DISALLOWED_TAGS = [
    "CUSTOM_ENCODED_XXE",
    "CUSTOM_ENCODED",
    "CUSTOM_VALIDATED_XXE",
    "CUSTOM_VALIDATED",
    "LIMITED_CHARS",
    "XMLIF_VALIDATED_XXE",
]

xxe_triggers = [
    {
        "module": "lxml.etree",
        "method_name": "fromstring",
        "source": "ARG_0,KWARG:text",
        "action": "FROMSTRING",
        "protect_mode": True,
    },
    {
        "module": "xml.dom.pulldom",
        "method_name": "parseString",
        "source": "ARG_0,KWARG:string",
        "protect_mode": True,
    },
    {
        "module": "xml.sax",
        "method_name": "parseString",
        "source": "ARG_0,KWARG:string",
        "protect_mode": True,
    },
]

register_trigger_rule(
    DataflowRule.from_nodes(
        "xxe",
        xxe_triggers,
        disallowed_tags=DISALLOWED_TAGS,
    )
)
